#include "libmpeg2.h"
#include <stdlib.h>


// Don't advance pointer
unsigned char mpeg2packet_next_char(mpeg2_demuxer_t *demuxer)
{
	return demuxer->raw_data[demuxer->raw_offset];
}

unsigned char mpeg2packet_read_char(mpeg2_demuxer_t *demuxer)
{
	unsigned char result = demuxer->raw_data[demuxer->raw_offset++];
	return result;
}

unsigned int mpeg2packet_read_int16(mpeg2_demuxer_t *demuxer)
{
	unsigned int a, b, result;
	a = demuxer->raw_data[demuxer->raw_offset++];
	b = demuxer->raw_data[demuxer->raw_offset++];
	result = (a << 8) | b;

	return result;
}

unsigned int mpeg2packet_next_int24(mpeg2_demuxer_t *demuxer)
{
	unsigned int a, b, c, result;
	a = demuxer->raw_data[demuxer->raw_offset];
	b = demuxer->raw_data[demuxer->raw_offset + 1];
	c = demuxer->raw_data[demuxer->raw_offset + 2];
	result = (a << 16) | (b << 8) | c;

	return result;
}

unsigned int mpeg2packet_read_int24(mpeg2_demuxer_t *demuxer)
{
	unsigned int a, b, c, result;
	a = demuxer->raw_data[demuxer->raw_offset++];
	b = demuxer->raw_data[demuxer->raw_offset++];
	c = demuxer->raw_data[demuxer->raw_offset++];
	result = (a << 16) | (b << 8) | c;

	return result;
}

unsigned int mpeg2packet_read_int32(mpeg2_demuxer_t *demuxer)
{
	unsigned int a, b, c, d, result;
	a = demuxer->raw_data[demuxer->raw_offset++];
	b = demuxer->raw_data[demuxer->raw_offset++];
	c = demuxer->raw_data[demuxer->raw_offset++];
	d = demuxer->raw_data[demuxer->raw_offset++];
	result = (a << 24) | (b << 16) | (c << 8) | d;

	return result;
}

unsigned int mpeg2packet_skip(mpeg2_demuxer_t *demuxer, long length)
{
	demuxer->raw_offset += length;
	return 0;
}

int mpeg2_get_adaptation_field(mpeg2_demuxer_t *demuxer)
{
	long length;
	int pcr_flag;

	demuxer->adaptation_fields++;
// get adaptation field length
	length = mpeg2packet_read_char(demuxer);
// get first byte
  	pcr_flag = (mpeg2packet_read_char(demuxer) >> 4) & 1;           

	if(pcr_flag)
	{
    	unsigned long clk_ref_base = mpeg2packet_read_int32(demuxer);
    	unsigned int clk_ref_ext = mpeg2packet_read_int16(demuxer);

		if (clk_ref_base > 0x7fffffff)
		{   // correct for invalid numbers
			clk_ref_base = 0;               // ie. longer than 32 bits when multiplied by 2
			clk_ref_ext = 0;                // multiplied by 2 corresponds to shift left 1 (<<=1)
		}
		else 
		{
			clk_ref_base <<= 1; // Create space for bit
			clk_ref_base |= (clk_ref_ext >> 15);          // Take bit
			clk_ref_ext &= 0x01ff;                        // Only lower 9 bits
		}
		demuxer->time = clk_ref_base + clk_ref_ext / 300;
	    if(length) mpeg2packet_skip(demuxer, length - 7);
	}
	else
	mpeg2packet_skip(demuxer, length - 1);

	return 0;
}

int mpeg2_get_program_association_table(mpeg2_demuxer_t *demuxer)
{
	demuxer->program_association_tables++;
	demuxer->table_id = mpeg2packet_read_char(demuxer);
	demuxer->section_length = mpeg2packet_read_int16(demuxer) & 0xfff;
	demuxer->transport_stream_id = mpeg2packet_read_int16(demuxer);
	mpeg2packet_skip(demuxer, demuxer->raw_size - demuxer->raw_offset);
	return 0;
}

int mpeg2packet_get_data_buffer(mpeg2_demuxer_t *demuxer)
{
	while(demuxer->raw_offset < demuxer->raw_size && demuxer->data_size < demuxer->data_allocated)
	{
		demuxer->data_buffer[demuxer->data_size++] = demuxer->raw_data[demuxer->raw_offset++];
	}
	return 0;
}

int mpeg2_get_pes_packet_header(mpeg2_demuxer_t *demuxer, unsigned long *pts, unsigned long *dts)
{
	unsigned int pes_header_bytes = 0;
	unsigned int pts_dts_flags;
	int pes_header_data_length;

// drop first 8 bits
	mpeg2packet_read_char(demuxer);  
	pts_dts_flags = (mpeg2packet_read_char(demuxer) >> 6) & 0x3;
	pes_header_data_length = mpeg2packet_read_char(demuxer);

// Get Presentation Time stamps and Decoding Time Stamps
	if(pts_dts_flags == 2)
	{
		*pts = (mpeg2packet_read_char(demuxer) >> 1) & 7;  // Only low 4 bits (7==1111)
		*pts <<= 15;
		*pts |= (mpeg2packet_read_int16(demuxer) >> 1);
		*pts <<= 15;
		*pts |= (mpeg2packet_read_int16(demuxer) >> 1);
		pes_header_bytes += 5;
	}
	else if(pts_dts_flags == 3)
	{      
		*pts = (mpeg2packet_read_char(demuxer) >> 1) & 7;  // Only low 4 bits (7==1111)
		*pts <<= 15;
		*pts |= (mpeg2packet_read_int16(demuxer) >> 1);
		*pts <<= 15;
		*pts |= (mpeg2packet_read_int16(demuxer) >> 1);
		*dts = (mpeg2packet_read_char(demuxer) >> 1) & 7;  // Only low 4 bits (7==1111)
		*dts <<= 15;
		*dts |= (mpeg2packet_read_int16(demuxer) >> 1);
		*dts <<= 15;
		*dts |= (mpeg2packet_read_int16(demuxer) >> 1);
		pes_header_bytes += 10;
  }
// extract other stuff here!
  
	mpeg2packet_skip(demuxer, pes_header_data_length - pes_header_bytes);
	return 0;
}

int get_unknown_data(mpeg2_demuxer_t *demuxer)
{
	mpeg2packet_skip(demuxer, demuxer->raw_size - demuxer->raw_offset);
	return 0;
}

int mpeg2_get_pes_packet_data(mpeg2_demuxer_t *demuxer, unsigned int stream_id)
{
	unsigned long pts = 0, dts = 0;

	if((stream_id >> 4) == 12 || (stream_id >> 4) == 13)
	{
// Just pick the first available stream if no ID is set
		if(demuxer->astream == -1)
		    demuxer->astream = (stream_id & 0x0f);

    	if((stream_id & 0x0f) == demuxer->astream && demuxer->do_audio)
		{
			mpeg2_get_pes_packet_header(demuxer, &pts, &dts);
			demuxer->pes_audio_time = pts;
			demuxer->audio_pid = demuxer->pid;
			demuxer->pes_audio_bytes += demuxer->raw_size - demuxer->raw_offset;
			return mpeg2packet_get_data_buffer(demuxer);
    	}
	}
	else 
	if((stream_id >> 4)==14)
	{
// Just pick the first available stream if no ID is set
		if(demuxer->vstream == -1)
			demuxer->vstream = (stream_id & 0x0f);

		if((stream_id & 0x0f) == demuxer->vstream && demuxer->do_video)
		{
			mpeg2_get_pes_packet_header(demuxer, &pts, &dts);
			demuxer->pes_video_time = pts;
			demuxer->video_pid = demuxer->pid;
			demuxer->pes_video_bytes += demuxer->raw_size - demuxer->raw_offset;
			return mpeg2packet_get_data_buffer(demuxer);
		}
	}
	else 
	{
		return get_unknown_data(demuxer);
	}

	mpeg2packet_skip(demuxer, demuxer->raw_size - demuxer->raw_offset);

	return 0;
}

int mpeg2_get_pes_packet(mpeg2_demuxer_t *demuxer)
{
	unsigned int stream_id;

	demuxer->pes_packets++;
	stream_id = mpeg2packet_read_char(demuxer);
// Skip startcode
	mpeg2packet_read_int24(demuxer);
// Skip pes packet length
	mpeg2packet_read_int16(demuxer);

	if(stream_id != MPEG2_PRIVATE_STREAM_2 && stream_id != MPEG2_PADDING_STREAM)
	{
		return mpeg2_get_pes_packet_data(demuxer, stream_id);
	}
	else
	if(stream_id == MPEG2_PRIVATE_STREAM_2)
	{
// Dump private data!
		fprintf(stderr, "stream_id == MPEG2_PRIVATE_STREAM_2\n");
		mpeg2packet_skip(demuxer, demuxer->raw_size - demuxer->raw_offset);
		return 0;
	}
	else
	if(stream_id == MPEG2_PADDING_STREAM)
	{
		mpeg2packet_skip(demuxer, demuxer->raw_size - demuxer->raw_offset);
		return 0;
	}
	else
	{
    	fprintf(stderr, "unknown stream_id in pes packet");
		return 1;
	}
	return 0;
}

int mpeg2_get_payload(mpeg2_demuxer_t *demuxer)
{
	if(demuxer->payload_unit_start_indicator)
	{
    	if(demuxer->pid==0) mpeg2_get_program_association_table(demuxer);
    	else 
		if(mpeg2packet_next_int24(demuxer) == MPEG2_PACKET_START_CODE_PREFIX) mpeg2_get_pes_packet(demuxer);
    	else 
		mpeg2packet_skip(demuxer, demuxer->raw_size - demuxer->raw_offset);
	}
	else
	{
    	if(demuxer->pid == demuxer->audio_pid && demuxer->do_audio)
		{
			demuxer->pes_audio_bytes += demuxer->raw_size - demuxer->raw_offset;
			mpeg2packet_get_data_buffer(demuxer);
		}
    	else 
		if(demuxer->pid == demuxer->video_pid && demuxer->do_video)
		{
			demuxer->pes_video_bytes += demuxer->raw_size - demuxer->raw_offset;
			mpeg2packet_get_data_buffer(demuxer);
		}
    	else 
		mpeg2packet_skip(demuxer, demuxer->raw_size - demuxer->raw_offset);
	}
	return 0;
}

// Read a transport packet
int mpeg2_read_transport(mpeg2_t *file, mpeg2_demuxer_t *demuxer)
{
	int result = mpeg2io_read_data(demuxer->raw_data, demuxer->packet_size, &(demuxer->fs));
	unsigned int bits;
	int table_entry;

	demuxer->raw_size = demuxer->packet_size;
	demuxer->raw_offset = 0;
	if(result)
	{
		perror("mpeg2_read_transport");
		return 1;
	}

// Sync byte
	if(mpeg2packet_read_char(demuxer) != MPEG2_SYNC_BYTE)
	{
		fprintf(stderr, "mpeg2packet_read_char(demuxer) != MPEG2_SYNC_BYTE\n");
		return 1;
	}

	bits = 	mpeg2packet_read_int24(demuxer) & 0x0000ffff;
	demuxer->transport_error_indicator = bits >> 15;
	demuxer->payload_unit_start_indicator = (bits >> 14) & 1;
	demuxer->pid = bits & 0x00001fff;
	demuxer->transport_scrambling_control = (mpeg2packet_next_char(demuxer) >> 6) & 0x3;
	demuxer->adaptation_field_control = (mpeg2packet_next_char(demuxer) >> 4) & 0x3;
	demuxer->continuity_counter = (mpeg2packet_read_char(demuxer) & 0xf);

	if(demuxer->transport_error_indicator)
	{
		fprintf(stderr, "demuxer->transport_error_indicator\n");
		return 1;
	}
 
    if (demuxer->pid == 0x1fff)
	{
		demuxer->is_padding = 1;  // padding; just go to next
		return 0;
    }
	else
	{
		demuxer->is_padding = 0;
	}

// Get pid
	for(table_entry = 0, result = 0; table_entry < demuxer->total_pids; table_entry++)
	{
		if(demuxer->pid == demuxer->pid_table[table_entry])
		{
			result = 1;
			break;
		}
	}

// Not in pid table
	if(!result)
	{
		demuxer->pid_table[table_entry] = demuxer->pid;
		demuxer->continuity_counters[table_entry] = demuxer->continuity_counter;  // init
		demuxer->total_pids++;
	}
	result = 0;

// Check counters
    if(demuxer->pid != MPEG2_PROGRAM_ASSOCIATION_TABLE && 
		demuxer->pid != MPEG2_CONDITIONAL_ACCESS_TABLE &&
        (demuxer->adaptation_field_control == 1 || demuxer->adaptation_field_control == 3))
	{
		if(demuxer->continuity_counters[table_entry] != demuxer->continuity_counter)
		{
			fprintf(stderr, "demuxer->continuity_counters[table_entry] != demuxer->continuity_counter\n");
// Reset it
			demuxer->continuity_counters[table_entry] = demuxer->continuity_counter;
		}
		if(++(demuxer->continuity_counters[table_entry]) > 15) demuxer->continuity_counters[table_entry] = 0;
	}

    if(demuxer->adaptation_field_control == 2 || demuxer->adaptation_field_control == 3)
    	result = mpeg2_get_adaptation_field(demuxer);

    if(demuxer->adaptation_field_control == 1 || demuxer->adaptation_field_control == 3)
    	result = mpeg2_get_payload(demuxer);

	return result;
}

int mpeg2_get_system_header(mpeg2_demuxer_t *demuxer)
{
	int length = mpeg2packet_read_int16(demuxer);
	unsigned long clock_ref, clock_ref_ext;

	
	mpeg2packet_skip(demuxer, length);
	return 0;
}


unsigned long mpeg2_get_timestamp(mpeg2_demuxer_t *demuxer)
{
	unsigned long timestamp;
// Only low 4 bits (7==1111)
	timestamp = (mpeg2packet_read_char(demuxer) >> 1) & 7;  
	timestamp <<= 15;
	timestamp |= (mpeg2packet_read_int16(demuxer) >> 1);
	timestamp <<= 15;
	timestamp |= (mpeg2packet_read_int16(demuxer) >> 1);
	return timestamp;
}

int mpeg2_get_pack_header(mpeg2_demuxer_t *demuxer, unsigned int *header)
{
	unsigned long i, j;
	unsigned long clock_ref, clock_ref_ext;

// Get the time code
	if((mpeg2packet_next_char(demuxer) >> 4) == 2)
	{
// MPEG-1
			demuxer->time = (double)mpeg2_get_timestamp(demuxer) / 90000;
// Skip 3 bytes
			mpeg2packet_read_int24(demuxer);
	}
	else
	if(mpeg2packet_next_char(demuxer) & 0x40)
	{
		i = mpeg2packet_read_int32(demuxer);
		j = mpeg2packet_read_int16(demuxer);
		if(i & 0x40000000 || (i >> 28) == 2)
		{
    		clock_ref = ((i & 0x31000000) << 3);
    		clock_ref |= ((i & 0x03fff800) << 4);
    		clock_ref |= ((i & 0x000003ff) << 5);
    		clock_ref |= ((j & 0xf800) >> 11);
    		clock_ref_ext = (j >> 1) & 0x1ff;

   			demuxer->time = (double)(clock_ref + clock_ref_ext / 300) / 90000;

// Skip 3 bytes
			mpeg2packet_read_int24(demuxer);
			i = mpeg2packet_read_char(demuxer) & 0x7;
// stuffing
			mpeg2packet_skip(demuxer, i);  
		}
	}
	else
	{
		mpeg2packet_skip(demuxer, 2);
	}

	*header = mpeg2packet_read_int32(demuxer);
	if(*header == MPEG2_SYSTEM_START_CODE)
	{
    	mpeg2_get_system_header(demuxer);
    	*header = mpeg2packet_read_int32(demuxer);
	}
	return 0;
}

int mpeg2_get_ps_pes_packet(mpeg2_demuxer_t *demuxer, unsigned int *header)
{
	unsigned long pts = 0, dts = 0;
	int stream_id = *header & 0xff;
	int pes_packet_length = mpeg2packet_read_int16(demuxer);
	int i;

	if(stream_id != MPEG2_PRIVATE_STREAM_2 && 
		stream_id != MPEG2_PADDING_STREAM)
	{
		if((mpeg2packet_next_char(demuxer) >> 6) == 0x02)
		{
// Get MPEG-2 packet
			int pes_header_bytes = 0;
			int scrambling = (mpeg2packet_read_char(demuxer) >> 4) & 0x3;
    		int pts_dts_flags = (mpeg2packet_read_char(demuxer) >> 6) & 0x3;
			int pes_header_data_length = mpeg2packet_read_char(demuxer);

			pes_packet_length -= 3;
// Get Presentation and Decoding Time Stamps
			if(pts_dts_flags == 2)
			{
				pts = (mpeg2packet_read_char(demuxer) >> 1) & 7;  // Only low 4 bits (7==1111)
				pts <<= 15;
				pts |= (mpeg2packet_read_int16(demuxer) >> 1);
				pts <<= 15;
				pts |= (mpeg2packet_read_int16(demuxer) >> 1);
				pes_header_bytes += 5;
				pes_packet_length -= 5;
			}
    		else 
			if(pts_dts_flags == 3)
			{
        		pts = (mpeg2packet_read_char(demuxer) >> 1) & 7;  // Only low 4 bits (7==1111)
        		pts <<= 15;
        		pts |= (mpeg2packet_read_int16(demuxer) >> 1);
        		pts <<= 15;
        		pts |= (mpeg2packet_read_int16(demuxer) >> 1);
        		dts = (mpeg2packet_read_char(demuxer) >> 1) & 7;  // Only low 4 bits (7==1111)
        		dts <<= 15;
        		dts |= (mpeg2packet_read_int16(demuxer) >> 1);
        		dts <<= 15;
        		dts |= (mpeg2packet_read_int16(demuxer) >> 1);
        		pes_header_bytes += 10;
				pes_packet_length -= 10;
    		}

// Skip unknown
        	mpeg2packet_skip(demuxer, pes_header_data_length - pes_header_bytes);
			pes_packet_length -= pes_header_data_length - pes_header_bytes;
		}
		else
    	if((mpeg2packet_next_char(demuxer) >> 6) == 0x01)
		{
// Skip header data entirely
       		mpeg2packet_skip(demuxer, 12);
      		pes_packet_length -= 12;
    	}
		else
		{
			int pts_dts_flags;
// Get MPEG-1 packet
			while(mpeg2packet_next_char(demuxer) == 0xff)
			{
				mpeg2packet_read_char(demuxer);
				pes_packet_length--;
			}

// Skip STD buffer scale
			if(mpeg2packet_next_char(demuxer) >= 0x40)
			{
				mpeg2packet_skip(demuxer, 2);
				pes_packet_length -= 2;
			}

// Decide which timestamps are available
			pts_dts_flags = mpeg2packet_next_char(demuxer);
			if(pts_dts_flags >= 0x30)
			{
// Get the presentation and decoding time stamp
				pts = mpeg2_get_timestamp(demuxer);
				dts = mpeg2_get_timestamp(demuxer);
				pes_packet_length -= 10;
			}
			else
			if(pts_dts_flags >= 0x20)
			{
// Get just the presentation time stamp
				pts = mpeg2_get_timestamp(demuxer);
				pes_packet_length -= 5;
			}
			else
			if(pts_dts_flags == 0x0f)
			{
// End of timestamps
				mpeg2packet_read_char(demuxer);
				pes_packet_length--;
			}
			else
			{
				return 1;     // Error
			}
		}

// Now extract the payload.
		if((stream_id >> 4) == 12 || (stream_id >> 4) == 13)
		{
// Audio data
// Take first stream ID if -1
			if(demuxer->astream == -1) demuxer->astream = stream_id & 0x0f;
			if((stream_id & 0x0f) == demuxer->astream && demuxer->do_audio)
			{
    			if(demuxer->pes_audio_bytes && pts)
				{
    				demuxer->pes_audio_time = 0;
    				demuxer->pes_audio_bytes = 0;
    			}

				if(pts) demuxer->pes_audio_time = pts;
   				demuxer->pes_audio_bytes += pes_packet_length;

				for(i = 0; i < pes_packet_length && demuxer->data_size < demuxer->data_allocated; i++)
				{
					demuxer->data_buffer[demuxer->data_size++] = demuxer->raw_data[demuxer->raw_offset++];
				}
		  	}
			else 
			{
    			mpeg2packet_skip(demuxer, demuxer->raw_size - demuxer->raw_offset);
			}
		}
    	else 
		if((stream_id >> 4) == 14)
		{
// Video data
// Take first stream ID if -1
			if(demuxer->vstream == -1) demuxer->vstream = stream_id & 0x0f;
    	    if((stream_id & 0x0f) == demuxer->vstream && demuxer->do_video)
			{
        		if(demuxer->pes_video_bytes && pts)
				{
        			demuxer->pes_video_time = 0;
        			demuxer->pes_video_bytes = 0;
        		}
				
        		if(pts) demuxer->pes_video_time = pts;
        		demuxer->pes_video_bytes += pes_packet_length;
//printf("mpeg2_get_ps_pes_packet 2 %d\n", pes_packet_length);
				for(i = 0; i < pes_packet_length && demuxer->data_size < demuxer->data_allocated; i++)
				{
					demuxer->data_buffer[demuxer->data_size++] = demuxer->raw_data[demuxer->raw_offset++];
				}
    	  	}
    		else 
			{
        	    mpeg2packet_skip(demuxer, demuxer->raw_size - demuxer->raw_offset);
    		}
    	}
    	else 
		if(stream_id == 0xbd)
		{
// AC 3 data
// Take first stream ID if -1
			if(demuxer->astream == -1) demuxer->astream = demuxer->raw_data[demuxer->raw_offset] - 0x80;

      		if((demuxer->raw_data[demuxer->raw_offset] == (0x80 + demuxer->astream)) && 
				demuxer->do_audio)
			{
        		if(demuxer->pes_audio_bytes && pts)
				{
        		    demuxer->pes_audio_time = 0;
        		    demuxer->pes_audio_bytes = 0;
        		}

        		if(pts) demuxer->pes_audio_time = pts;
        		demuxer->pes_audio_bytes += pes_packet_length - 4;
				mpeg2packet_read_int32(demuxer);
				for(i = 0; i < (pes_packet_length - 4) && demuxer->data_size < demuxer->data_allocated; i++)
				{
					demuxer->data_buffer[demuxer->data_size++] = demuxer->raw_data[demuxer->raw_offset++];
				}
      		}
      		else 
			{
        	    mpeg2packet_skip(demuxer, demuxer->raw_size - demuxer->raw_offset);
      		}
    	}
    	else 
		if(stream_id == 0xbc || 1)
		{
        	mpeg2packet_skip(demuxer, demuxer->raw_size - demuxer->raw_offset);
    	}
	}
  	else 
	if(stream_id == MPEG2_PRIVATE_STREAM_2)
	{
        mpeg2packet_skip(demuxer, demuxer->raw_size - demuxer->raw_offset);
  	}
  	else 
	if(stream_id == MPEG2_PADDING_STREAM)
	{
        mpeg2packet_skip(demuxer, demuxer->raw_size - demuxer->raw_offset);
  	}

    mpeg2packet_skip(demuxer, demuxer->raw_size - demuxer->raw_offset);
	return 0;
}

int mpeg2_read_program(mpeg2_t *file, mpeg2_demuxer_t *demuxer)
{
	int result = 0;
	unsigned int header;
	demuxer->raw_size = demuxer->packet_size;
	demuxer->raw_offset = 0;
	demuxer->data_size = 0;

// Search backward for it.
	header = mpeg2io_read_int32(&(demuxer->fs));
	result |= mpeg2io_eof(&(demuxer->fs));
	result |= mpeg2io_seek_relative(&(demuxer->fs), -4);

	while(header != MPEG2_PACK_START_CODE && !result)
	{
		result |= mpeg2io_seek_relative(&(demuxer->fs), -1);
		header >>= 8;
		header |= mpeg2io_read_char(&(demuxer->fs)) << 24;
		result |= mpeg2io_seek_relative(&(demuxer->fs), -1);
	}

	if(result)
	{
//		fprintf(stderr, "mpeg2_read_program: couldn't find MPEG2_PACK_START_CODE\n");
		return 1;
	}

	result |= mpeg2io_read_data(demuxer->raw_data, demuxer->packet_size, &(demuxer->fs));

	if(result)
	{
		perror("mpeg2_read_transport");
		return 1;
	}

	header = mpeg2packet_read_int32(demuxer);
	while(demuxer->raw_offset < demuxer->raw_size)
	{
		if(header == MPEG2_PACK_START_CODE)
		{
			result |= mpeg2_get_pack_header(demuxer, &header);
		}
		else
		if((header >> 8) == MPEG2_PACKET_START_CODE_PREFIX)
		{
			result |= mpeg2_get_ps_pes_packet(demuxer, &header);
		}
	}
	return result;
}

// Read packet in the forward direction
int mpeg2_read_next_packet(mpeg2_t *file, mpeg2_demuxer_t *demuxer)
{
	int result = 0;
	long current_position;

	demuxer->data_size = 0;
	demuxer->data_position = 0;

// Flip the file descriptor back to the end of the packet for forward
// reading.
	if(demuxer->reverse)
	{
		result |= mpeg2io_seek_relative(&(demuxer->fs), demuxer->packet_size);
		demuxer->reverse = 0;
	}

	do
	{
		if(file->is_transport_stream && !result)
		{
			result |= mpeg2_read_transport(file, demuxer);
		}
		else
		if(file->is_program_stream && !result)
		{
			result |= mpeg2_read_program(file, demuxer);
		}
		else
		if(!result)
		{
	// Read elementary stream.
			result |= mpeg2io_read_data(demuxer->data_buffer, demuxer->packet_size, &(demuxer->fs));
			if(!result) demuxer->data_size = demuxer->packet_size;
		}
	}while(!result && demuxer->data_size == 0);

	demuxer->error_flag = result;
	return result;
}

// Read the packet right before the packet we're currently on.

int mpeg2_read_prev_packet(mpeg2_t *file, mpeg2_demuxer_t *demuxer)
{
	int result = 0;
	long current_position;

	demuxer->data_size = 0;
	demuxer->data_position = 0;

	do
	{
// Rewind to the start of the packet to be read.
		result |= mpeg2io_seek_relative(&(demuxer->fs), -demuxer->packet_size);

		if(file->is_transport_stream && !result)
		{
			result |= mpeg2_read_transport(file, demuxer);
			if(!result) result |= mpeg2io_seek_relative(&(demuxer->fs), -demuxer->packet_size);
		}
		else
		if(file->is_program_stream && !result)
		{
			result |= mpeg2_read_program(file, demuxer);
			if(!result) result |= mpeg2io_seek_relative(&(demuxer->fs), -demuxer->packet_size);
		}
		else
		{
// Elementary stream
// Read the packet forwards and seek back to the start
			result |= mpeg2io_read_data(demuxer->data_buffer, demuxer->packet_size, &(demuxer->fs));
			if(!result) 
			{
				demuxer->data_size = demuxer->packet_size;
				result |= mpeg2io_seek_relative(&(demuxer->fs), -demuxer->packet_size);
			}
		}
	}while(!result && demuxer->data_size == 0);

// Remember that the file descriptor is at the beginning of the packet just read.
	demuxer->reverse = 1;
	demuxer->error_flag = result;
	return result;
}


// ====================================================================
//                            Entry points
// ====================================================================

mpeg2_demuxer_t* mpeg2_new_demuxer(mpeg2_t *file, int do_audio, int do_video)
{
	mpeg2_demuxer_t *demuxer = malloc(sizeof(mpeg2_demuxer_t));
	unsigned long test_header = 0;
	int i;

// The demuxer can change the default packet size for its own use.
	demuxer->packet_size = file->packet_size;
	demuxer->do_audio = do_audio;
	demuxer->do_video = do_video;
	demuxer->data_buffer = 0;
	demuxer->data_size = 0;
	demuxer->data_allocated = 0;
	demuxer->data_position = 0;
	demuxer->reverse = 0;
	demuxer->error_flag = 0;

	if(mpeg2io_open_file(&(demuxer->fs), file->fs.path))
	{
		demuxer->error_flag = 1;
		perror("mpeg2_new_demuxer");
	}
	
// Get the packet size from the file
	if(file->is_program_stream)
	{
		mpeg2io_seek(&(demuxer->fs), 4);
		for(i = 0; i < 256000 && test_header != MPEG2_PACK_START_CODE; i++)
		{
			test_header <<= 8;
			test_header |= mpeg2io_read_char(&(demuxer->fs));
		}
		if(i < 256000) demuxer->packet_size = i;
		mpeg2io_seek(&(demuxer->fs), 0);
	}

// Allocate buffer
	demuxer->raw_data = malloc(demuxer->packet_size);
	demuxer->raw_offset = 0;
	demuxer->data_buffer = malloc(file->packet_size);
	demuxer->data_size = 0;
	demuxer->data_allocated = file->packet_size;
// System specific variables
	demuxer->is_padding = 0;
	demuxer->total_pids = 0;
	demuxer->adaptation_fields = 0;
	demuxer->time = 0;
	demuxer->transport_error_indicator = 0;
	demuxer->payload_unit_start_indicator = 0;
	demuxer->pid = 0;
	demuxer->transport_scrambling_control = 0;
	demuxer->adaptation_field_control = 0;
	demuxer->continuity_counter = 0;
	demuxer->audio_pid = -1;
	demuxer->video_pid = -1;
	demuxer->astream = -1;
	demuxer->vstream = -1;
	demuxer->program_association_tables = 0;
	demuxer->table_id = 0;
	demuxer->section_length = 0;
	demuxer->transport_stream_id = 0;
	demuxer->pes_packets = 0;
	demuxer->pes_audio_time = 0;
	demuxer->pes_video_time = 0;
	demuxer->pes_audio_bytes = 0;
	demuxer->pes_video_bytes = 0;
	return demuxer;
}

int mpeg2_delete_demuxer(mpeg2_t *file, mpeg2_demuxer_t *demuxer)
{
	mpeg2io_close_file(&(demuxer->fs));
	if(demuxer->data_buffer) free(demuxer->data_buffer);
	free(demuxer->raw_data);
	free(demuxer);
}

int mpeg2demux_read_data(mpeg2_t *file, mpeg2_demuxer_t *demuxer, unsigned char *output, long size)
{
	long i;
	int result = 0;
	
	if(demuxer->data_position >= 0)
	{
// Read forwards
		for(i = 0; i < size && !result; )
		{
			while(i < size && demuxer->data_position < demuxer->data_size)
			{
				output[i++] = demuxer->data_buffer[demuxer->data_position++];
			}

			if(i < size)
			{
				result |= mpeg2_read_next_packet(file, demuxer);
			}
		}
	}
	else
	{
// Read backwards a full packet.
// Only good for reading less than the size of a full packet, but
// this routine should only be used for searching for previous markers.
		long current_position = demuxer->data_position;
		result |= mpeg2_read_prev_packet(file, demuxer);
		if(!result) demuxer->data_position = demuxer->data_size + current_position;
		for(i = 0; i < size; )
		{
			output[i++] = demuxer->data_buffer[demuxer->data_position++];
		}
	}

	demuxer->error_flag = result;
	return result;
}

int mpeg2demux_seek(mpeg2_t *file, mpeg2_demuxer_t *demuxer, long byte)
{
	long current_position;
	demuxer->data_position = 0;
	demuxer->data_size = 0;

	demuxer->error_flag = mpeg2io_seek(&(demuxer->fs), byte);

	if(!demuxer->error_flag && (file->is_transport_stream || file->is_program_stream))
	{
// Get on a packet boundary only for system streams.
		current_position = mpeg2io_tell(&(demuxer->fs));
		if(byte % demuxer->packet_size)
		{
			demuxer->error_flag |= mpeg2io_seek(&(demuxer->fs), current_position - (current_position % demuxer->packet_size));
		}
	}
	return demuxer->error_flag;
}

int mpeg2demux_error(mpeg2_t *file, mpeg2_demuxer_t *demuxer)
{
	return demuxer->error_flag;
}

unsigned int mpeg2demux_read_char(mpeg2_t *file, mpeg2_demuxer_t *demuxer)
{
	unsigned char a;
	demuxer->error_flag = mpeg2demux_read_data(file, demuxer, &a, 1);
	return a;
}

unsigned int mpeg2demux_read_prev_char(mpeg2_t *file, mpeg2_demuxer_t *demuxer)
{
	unsigned char a;
	demuxer->data_position--;
	demuxer->error_flag = mpeg2demux_read_data(file, demuxer, &a, 1);
	demuxer->data_position--;
	return a;
}
