#include "aedit.h"
#include "aedits.h"
#include "assets.h"
#include "atrack.h"
#include "autoconf.h"
#include "edit.h"
#include "cache.h"
#include "datatype.h"
#include "file.h"
#include "filehtal.h"
#include "floatautos.h"
#include "mainwindow.h"
#include "pluginbuffer.h"
#include "trackcanvas.h"
#include "tracks.h"

ATrack::ATrack(MainWindow *mwindow, Tracks *tracks) : Track(mwindow, tracks)
{
	data_type = TRACK_AUDIO;
}

ATrack::~ATrack()
{
	delete edits;
	for(int i = 0; i < mwindow->output_channels; i++) delete pan_autos[i];
}

int ATrack::create_derived_objs(int flash)
{
//printf("ATrack::create_derived_objects 1\n");
	int i;
	edits = new AEdits(mwindow, this);

	for(i = 0; i < mwindow->output_channels; i++)
	{
		pan_autos[i] = new FloatAutos(this, MEPURPLE, -1, 1, PAN_AUTO_H, 1);
		pan_autos[i]->create_objects();
	}
	
	fade_autos = new FloatAutos(this, LTGREY, -46, 46);
	fade_autos->create_objects();

	if(mwindow->gui) 
	{
		if(mwindow->tracks_vertical)
		draw(pixel, mwindow->zoom_track, 0, tracks->canvas->get_h(), flash);
		else
		draw(0, tracks->canvas->get_w(), pixel, mwindow->zoom_track, flash);
	}
}

int ATrack::set_index_files(int flash, Asset *asset)
{
	int result = 0;
	AEdit* current;

	if(mwindow->gui)
	{
		result = 1;
		for(current = (AEdit*)edits->first; current; current = (AEdit*)NEXT)
		{
// check for matching asset
			if(current->asset && current->asset == asset)
			{
				if(mwindow->tracks_vertical)
					current->set_index_file(flash, pixel + mwindow->zoom_track / 2, 0, 0, tracks->canvas->get_w(), tracks->canvas->get_h());
				else
					current->set_index_file(flash, pixel + mwindow->zoom_track / 2, 0, 0, tracks->canvas->get_w(), tracks->canvas->get_h());

				result = 0;
			}
		}
	}

	return result;
}


int ATrack::save_derived(FileHTAL *htal)
{
	int i;
	edits->save(htal);

	
	for(i = 0; i < mwindow->output_channels; i++)
	{
		if(pan_autos[i]->total())
		{
			htal->tag.set_title("PANAUTOS");
			htal->tag.set_property("CHANNEL", i);
			htal->append_tag();
			pan_autos[i]->save(htal);
			htal->tag.set_title("/PANAUTOS");
			htal->append_tag();
			htal->append_newline();
		}
	}
}

int ATrack::load_derived(FileHTAL *htal, int automation_only, int edits_only, int load_all, int &current_channel)
{
	if(htal->tag.title_is("PANAUTOS"))
	{
		current_channel = htal->tag.get_property("CHANNEL", current_channel);
		if(current_channel < mwindow->output_channels) pan_autos[current_channel++]->load(htal, "/PANAUTOS");
	}
}

int ATrack::change_channels(int oldchannels, int newchannels)
{
	int i;
	
	if(oldchannels > newchannels)
// delete extra channels
	{
		for(i = newchannels; i < oldchannels; i++)
		{
			delete(pan_autos[i]);
		}
	}
	else
// add new channels
	{
		for(i = oldchannels; i < newchannels; i++)
		{
			pan_autos[i] = new FloatAutos(this, MEPURPLE, -1, 1, PAN_AUTO_H, 1);
			pan_autos[i]->create_objects();
		}
	}
}

int ATrack::render(PluginBuffer *shared_output, 
			   long offset, long input_len, 
               long input_position) // always start of range
{
	long input_start = input_position;		   // start of render in project
	long output_start = 0;  	// start position in output buffer of current edit
	long render_len = input_len;			// length of render in project
	long input_end = input_position + render_len; 			// end of render in project
	long output_end = render_len;			 // end of buffer
	AEdit *current_edit;

// zero buffer
	for(register int i = output_start; i < output_end; i++) 
	{ 
//		output[i] = 0; 
		*((float*)shared_output->get_data() + offset + i) = 0; 
	}

// get first edit contained in range for forward render
	for(current_edit = (AEdit*)edits->first; 
		current_edit && current_edit->startproject + current_edit->length + current_edit->feather_right <= input_start;
		current_edit = (AEdit*)current_edit->next)
	{ ; }

// render edits
	while(current_edit && current_edit->startproject - current_edit->feather_left < input_end)
	{
// test for asset
		if(!current_edit->asset)
		{   // skip the edit
			current_edit = (AEdit*)current_edit->next;
		}
		else
// test for silence
		if(current_edit->asset->silence)      
		{   // skip the edit
			current_edit = (AEdit*)current_edit->next;
		}
		else
// edit contains data
		{
			File *source;
//printf("ATrack::render checking out\n");
			if(!(source = mwindow->cache->check_out(current_edit->asset)))
			{
// couldn't open source file / skip the edit
				printf("ATrack::render Couldn't open %s.\n", current_edit->asset->path);
				current_edit = (AEdit*)current_edit->next;
			}
			else
			{ 
// render the edit
// add the feather range
				long lfeather_start, lfeather_len;
				float lfeather_slope, lfeather_gain;
				long rfeather_start, rfeather_len;
				float rfeather_slope, rfeather_gain;
				
// get the left feather range and slope
				lfeather_len = current_edit->feather_left * 2;
				lfeather_start = current_edit->startproject - current_edit->feather_left;
				if(lfeather_len)
				{
					lfeather_slope = (float)1 / lfeather_len;
					lfeather_gain = 0;
				}

// get the right feather range and slope
				rfeather_len = current_edit->feather_right * 2;
				rfeather_start = current_edit->startproject + current_edit->length - current_edit->feather_right;
				if(rfeather_len)
				{
					rfeather_slope = -(float)1 / rfeather_len;
					rfeather_gain = 1;
				}

// test if left and right feathers overlap
				if(lfeather_start + lfeather_len > rfeather_start)
				{
					long center = (long)((rfeather_gain - lfeather_gain) / (lfeather_slope - rfeather_slope));
					rfeather_len -= center - rfeather_start;             // decrease the length
					rfeather_gain += rfeather_slope * (center - rfeather_start);   // decrease the gain
					lfeather_len -= center - rfeather_start;              // decrease the length
					rfeather_start = center;                          // change the start point
				}
				
// test feather boundaries
// clear feathers that are entirely outside of the range
				if(lfeather_start + lfeather_len <= input_start) 
				{           // zero out the left  feather
					lfeather_start = input_start;
					lfeather_len = 0;  
				}

				if(rfeather_start >= input_end)
				{           // zero out the right feather
					rfeather_start = input_end;
					rfeather_len = 0;
				}

// adjust feathers that are partly in range
// left feather starts before range
				if(lfeather_start < input_start)
				{
					lfeather_gain += (input_start - lfeather_start) * lfeather_slope;
					lfeather_len -= input_start - lfeather_start;
					lfeather_start = input_start;
				}

// left feather ends after range
				if(lfeather_start + lfeather_len > input_end)
				{
					lfeather_len = input_end - lfeather_start;
				}

// right feather starts before range
				if(rfeather_start < input_start)
				{
					rfeather_gain += rfeather_slope * (input_start - rfeather_start);   // decrease the initial gain
					rfeather_len -= (input_start - rfeather_start);
					rfeather_start = input_start;
				}

// right feather ends after range
				if(rfeather_start + rfeather_len > input_end)
				{
					rfeather_len = input_end - rfeather_start;
				}
				

// get source file positions
				long startsource, lengthsource;
				startsource = current_edit->startsource + lfeather_start - current_edit->startproject;

// test source boundaries
// source starts before file
				if(startsource < 0)
				{
					lfeather_start += -startsource;
					lfeather_len -= -startsource;
					if(lfeather_len < 0) lfeather_len = 0;
					lfeather_gain += -startsource * lfeather_slope;
					startsource = 0;
				}

// need lengthsource to determine if length exceeds length of file
				lengthsource = rfeather_start + rfeather_len - lfeather_start;
// source ends after end of file
				if(startsource + lengthsource > source->get_length())
				{
					rfeather_len -= startsource + lengthsource - source->get_length();
					if(rfeather_len < 0) rfeather_len = 0;
					lengthsource -= startsource + lengthsource - source->get_length();
				}

// set the start position in the output buffer of the edit
				output_start = lfeather_start - input_start;    // get the generic start

// read the data
				source->set_position(startsource);
				source->set_channel(current_edit->channel);
// 				source->render_samples(&output[output_start], lengthsource,
// 					lfeather_len, lfeather_gain, lfeather_slope,
// 					rfeather_len, rfeather_gain, rfeather_slope);
				source->render_samples(shared_output, 
					output_start + offset, 
					lengthsource,
					lfeather_len, lfeather_gain, lfeather_slope,
					rfeather_len, rfeather_gain, rfeather_slope);

				mwindow->cache->check_in(current_edit->asset);

// advance the edit
				current_edit = (AEdit*)current_edit->next;
			}

		}
	}
}

// =========================== drawing commands ==================

int ATrack::draw_derived(int x, int w, int y, int h, int flash)
{       // make sure this track is visible
	int center_pixel;    // pixel of center line in canvas
	center_pixel = pixel + mwindow->zoom_track / 2;

	tracks->canvas->set_color(RED);        // draw zero line
	if(mwindow->tracks_vertical)
	tracks->canvas->draw_line(center_pixel, y, center_pixel, y + h);
	else
	tracks->canvas->draw_line(x, center_pixel, x + w, center_pixel);
}


int ATrack::draw_autos_derived(float view_start, float zoom_units, AutoConf *auto_conf)
{
	int i;
	
	for(i = 0; i < mwindow->output_channels; i++)
	{
		if(auto_conf->pan[i]) 
			pan_autos[i]->draw(tracks->canvas, 
							pixel, 
							mwindow->zoom_track, 
							zoom_units, 
							view_start, 
							mwindow->tracks_vertical);
	}
}

int ATrack::draw_floating_autos_derived(float view_start, float zoom_units, AutoConf *auto_conf, int flash)
{
	int i;

	for(i = 0; i < mwindow->output_channels; i++)
	{
		if(auto_conf->pan[i]) 
			pan_autos[i]->draw_floating_autos(tracks->canvas, 
									pixel, 
									mwindow->zoom_track, 
									zoom_units, 
									view_start, 
									mwindow->tracks_vertical, flash);
	}
}

long ATrack::length()
{
	return edits->end();
}

int ATrack::get_dimensions(float &view_start, float &view_units, float &zoom_units)
{
	view_start = (float)mwindow->view_start;
	view_units = (float)tracks->view_samples();
	zoom_units = (float)mwindow->zoom_sample;
}

long ATrack::samples_to_units(long &samples)
{
	return samples;
}


long ATrack::units_to_samples(long &units)
{
	return units;
}

int ATrack::copy_derived(long start, long end, FileHTAL *htal)
{
// automation is taken care of by Track::copy_automation
}

int ATrack::copy_automation_derived(AutoConf *auto_conf, long selectionstart, long selectionend, FileHTAL *htal)
{
	int i, result;

	for(i = 0, result = 0; i < mwindow->output_channels; i++)
	{
		if(auto_conf->pan[i])
		{
			htal->tag.set_title("PANAUTOS");
			htal->tag.set_property("CHANNEL", i);
			htal->append_tag();

			pan_autos[i]->copy(selectionstart, selectionend, htal, 1);

			htal->tag.set_title("/PANAUTOS");
			htal->append_tag();
			htal->append_newline();
			result = 1;
		}
	}

	return result;
}


int ATrack::paste_automation_derived(long selectionstart, long selectionend, long total_length, FileHTAL *htal, int shift_autos, int &current_pan)
{
// only used for automation editing routines
	if(htal->tag.title_is("PANAUTOS") && current_pan < MAXCHANNELS)
	{
		current_pan = htal->tag.get_property("CHANNEL", current_pan);
		pan_autos[current_pan++]->paste(selectionstart, selectionend, total_length, htal, "/PANAUTOS", 1, shift_autos);
	}
	return 1;
}

int ATrack::clear_automation_derived(AutoConf *auto_conf, long selectionstart, long selectionend, int shift_autos)
{
// used when clearing just automation
	int i, result;

	for(i = 0, result = 0; i < mwindow->output_channels && !result; i++)
		if(auto_conf->pan[i]) result = 1;

	for(i = 0; i < mwindow->output_channels && result; i++)
	{
		pan_autos[i]->clear(selectionstart, selectionend, 1, shift_autos);
	}
	
	return result;
}

int ATrack::paste_derived(long start, long end, long total_length, FileHTAL *htal, int &current_channel)
{
	if(!strcmp(htal->tag.get_title(), "PANAUTOS"))
	{
		current_channel = htal->tag.get_property("CHANNEL", current_channel);

		if(current_channel < mwindow->output_channels) 
			pan_autos[current_channel++]->paste(start, end, total_length, htal, "/PANAUTOS", mwindow->autos_follow_edits);
		return 1;
	}
	return 0;
}

int ATrack::paste_output(long startproject, long endproject, long startsource, long endsource, int channel, Asset *asset)
{
	int result = 0;

	result = ((AEdits*)edits)->paste_edit(startproject, endproject, startsource, endsource - startsource, channel, asset);
	//if(!result && mwindow->autos_follow_edits)
	//{
	//	paste_auto_silence(startproject, endproject);
	//}
	return result;
}

int ATrack::clear_derived(long start, long end)
{
// used when clearing everything
	int i;
	if(mwindow->autos_follow_edits)
	{
		for(i = 0; i < mwindow->output_channels; i++)
		{
			pan_autos[i]->clear(start, end, 1, 1);
		}
	}
}

int ATrack::paste_auto_silence_derived(long start, long end)
{
	for(int i = 0; i < mwindow->output_channels; i++)
	{
		pan_autos[i]->paste_silence(start, end);
	}
}

int ATrack::select_auto_derived(float zoom_units, float view_start, AutoConf *auto_conf, int cursor_x, int cursor_y)
{
	int result = 0;
	int i, temp;

	for(i = mwindow->output_channels - 1, result = 0; i >= 0 && !result; i--)
	{
		if(auto_conf->pan[i]) 
			result = pan_autos[i]->select_auto(tracks->canvas, 
											pixel, 
											mwindow->zoom_track, 
											zoom_units, 
											view_start, 
											cursor_x, 
											cursor_y, 
											mwindow->tracks_vertical);
	}
	
	return result;
}

int ATrack::move_auto_derived(float zoom_units, float view_start, AutoConf *auto_conf, int cursor_x, int cursor_y, int shift_down)
{
	int result, i;
	result = 0;
	
	for(i = 0; i < mwindow->output_channels && !result; i++)
	{
		if(auto_conf->pan[i]) 
			result = pan_autos[i]->move_auto(tracks->canvas, 
										pixel, 
										mwindow->zoom_track, 
										zoom_units, 
										view_start, 
										cursor_x, 
										cursor_y, 
										shift_down, 
										mwindow->tracks_vertical);
	}
		
	return result;
}

int ATrack::release_auto_derived()
{
	int result, i;
	result = 0;

	for(i = 0; i < mwindow->output_channels && !result; i++)
	{
		result = pan_autos[i]->release_auto();
	}
		
	return result;
}

int ATrack::scale_time_derived(float rate_scale, int scale_edits, int scale_autos, long start, long end)
{
	for(int i = 0; i < mwindow->output_channels; i++)
	{
		pan_autos[i]->scale_time(rate_scale, scale_edits, scale_autos, start, end);
	}
}

