#include "audiodevice.h"
#include "dcoffset.h"


AudioDevice::AudioDevice()
 : Thread()
{
// get better synchronization by starting playback first
	RECORDBEFOREPLAY = 0;
	
	r = w = d = 0;
	dsp_in = dsp_out = dsp_duplex = 0;
	ochannels = ichannels = dchannels = 0;
	orate = irate = drate = 0;
	obits = ibits = dbits = 0;

	for(int i = 0; i < TOTAL_BUFFERS; i++)
	{
		buffer[i] = 0;
		buffer_size[i] = 0;
		play_mutex[i].lock();
		last_buffer[i] = 0;
	}

	input_buffer = 0;

	duplex_init = 0;
	rec_dither = play_dither = 0;
	software_position_info = 0;
	arm_buffer_num = 0;
	is_playing_back = 0;
	is_recording = 0;
	last_buffer_size = total_samples = position_correction = 0;
	last_position = 0;
	dc_offset_thread = 0;
	interrupt = 0;
}

AudioDevice::~AudioDevice()
{
}

long AudioDevice::bits_to_fmt(int bits)
{
#ifndef OTHERAUDIO
	switch(bits)
	{
		case 8:
			return AFMT_S8;
			break;
			
// intel formats
		case 16:
			return AFMT_S16_LE;
			break;
			
		case 24:
			return AFMT_S24_LE;
			break;
	}
#endif
	return 0;
}

int AudioDevice::open_input(char *idevice, int bits, int rate, int channels, long size)
{
	int format = bits_to_fmt(bits);
	int buffer_info;
	duplex_init = 0;

	buffer_info = sizetofrag(size, channels, bits);

	dsp_in = open(idevice, O_RDONLY/* | O_NDELAY*/);
	if(dsp_in < 0) printf("open odevice failed\n");
	set_cloexec_flag(dsp_in, 1);

#ifndef OTHERAUDIO
	if(ioctl(dsp_in, SNDCTL_DSP_SETFRAGMENT, &buffer_info)) printf("SNDCTL_DSP_SETFRAGMENT failed.\n");
	if(ioctl(dsp_in, SNDCTL_DSP_SETFMT, &format) < 0) printf("SNDCTL_DSP_SETFMT failed\n");
	if(ioctl(dsp_in, SNDCTL_DSP_CHANNELS, &channels) < 0) printf("SNDCTL_DSP_CHANNELS failed\n");
	if(ioctl(dsp_in, SNDCTL_DSP_SPEED, &rate) < 0) printf("SNDCTL_DSP_SPEED failed\n");
#endif
	r = 1;
	strcpy(this->idevice, idevice);
	ichannels = channels;
	ibits = bits;
	irate = rate;
	dc_offset_thread = new DC_Offset;
//printf("AudioDevice::open_input %x\n", dc_offset_thread);
}

int AudioDevice::open_output(char *odevice, int bits, int rate, int channels, int size)
{
	duplex_init = 0;
	int format = bits_to_fmt(bits);
	int buffer_info;

	buffer_info = sizetofrag(size, channels, bits);

	dsp_out = open(odevice, O_WRONLY /*| O_NDELAY*/);
	if(dsp_out < 0) printf("open odevice failed\n");
	set_cloexec_flag(dsp_out, 1);

#ifndef OTHERAUDIO
	audio_buf_info playinfo;

	if(ioctl(dsp_out, SNDCTL_DSP_SETFRAGMENT, &buffer_info)) printf("SNDCTL_DSP_SETFRAGMENT failed.\n");
	if(ioctl(dsp_out, SNDCTL_DSP_SETFMT, &format) < 0) printf("SNDCTL_DSP_SETFMT failed\n");
	if(ioctl(dsp_out, SNDCTL_DSP_CHANNELS, &channels) < 0) printf("SNDCTL_DSP_CHANNELS failed\n");
	if(ioctl(dsp_out, SNDCTL_DSP_SPEED, &rate) < 0) printf("SNDCTL_DSP_SPEED failed\n");
	ioctl(dsp_out, SNDCTL_DSP_GETOSPACE, &playinfo);
	device_buffer = playinfo.bytes;
#endif
	w = 1;
	strcpy(this->odevice, odevice);
	ochannels = channels;
	obits = bits;
	orate = rate;
}

int AudioDevice::open_duplex(char *device, int bits, int rate, int channels, long size)
{
	int format = bits_to_fmt(bits);
	int buffer_info = sizetofrag(size, channels, bits);
	//int trigger_bits = ~PCM_ENABLE_OUTPUT || ~PCM_ENABLE_INPUT;

	dsp_duplex = open(device, O_RDWR/* | O_NDELAY*/);
	if(dsp_duplex < 0) printf("open odevice failed\n");
	set_cloexec_flag(dsp_duplex, 1);

#ifndef OTHERAUDIO
	audio_buf_info playinfo;

	if(ioctl(dsp_duplex, SNDCTL_DSP_SETFRAGMENT, &buffer_info)) printf("SNDCTL_DSP_SETFRAGMENT failed.\n");
	if(ioctl(dsp_duplex, SNDCTL_DSP_SETDUPLEX, 1) == -1) printf("SNDCTL_DSP_SETDUPLEX failed\n");
	if(ioctl(dsp_duplex, SNDCTL_DSP_SETFMT, &format) < 0) printf("SNDCTL_DSP_SETFMT failed\n");
	if(ioctl(dsp_duplex, SNDCTL_DSP_CHANNELS, &channels) < 0) printf("SNDCTL_DSP_CHANNELS failed\n");
	if(ioctl(dsp_duplex, SNDCTL_DSP_SPEED, &rate) < 0) printf("SNDCTL_DSP_SPEED failed\n");
	ioctl(dsp_duplex, SNDCTL_DSP_GETOSPACE, &playinfo);
	device_buffer = playinfo.bytes;

	//if(ioctl(dsp_duplex, SNDCTL_DSP_SETTRIGGER, &trigger_bits) < 0) printf("SNDCTL_DSP_SETTRIGGER failed\n");;
#endif
	d = 1;
	duplex_init = 1;        // notify playback routines to test the duplex lock
	duplex_lock.lock();     // prevent playback until recording starts

	strcpy(this->ddevice, device);
	dchannels = channels;
	dbits = bits;
	drate = rate;
	dc_offset_thread = new DC_Offset;
}





int AudioDevice::close_all()
{
#ifndef OTHERAUDIO
	if(r) { ioctl(dsp_in, SNDCTL_DSP_RESET, 0);         close(dsp_in);      r = 0; }
	if(w) { ioctl(dsp_out, SNDCTL_DSP_RESET, 0);        close(dsp_out);     w = 0; /*printf("close(dsp_out)\n");*/}
	if(d) { ioctl(dsp_duplex, SNDCTL_DSP_RESET, 0);     close(dsp_duplex);  d = 0; }
#endif

	if(dc_offset_thread) { delete dc_offset_thread;   dc_offset_thread = 0; }
	reset_output();
	if(input_buffer) { delete input_buffer; input_buffer = 0; }
	is_recording = 0;
	rec_dither = play_dither = 0;
	software_position_info = position_correction = last_buffer_size = 0;
	r = w = d = 0;
	duplex_init = 0;
}

int AudioDevice::set_cloexec_flag(int desc, int value)
{
	int oldflags = fcntl (desc, F_GETFD, 0);
	if (oldflags < 0) return oldflags;
	if(value != 0) 
		oldflags |= FD_CLOEXEC;
	else
		oldflags &= ~FD_CLOEXEC;
	return fcntl(desc, F_SETFD, oldflags);
}



int AudioDevice::sizetofrag(int size, int channels, int bits)
{
	int testfrag = 2, fragsize = 1;
	size *= channels * bits / 8;
	while(testfrag < size)
	{
		fragsize++;
		testfrag *= 2;
	}
	return (4<<16) | fragsize;
}
