#include "assets.h"
#include "file.h"
#include "filempeg.h"
#include "vframe.h"



FileMPEG::FileMPEG(Asset *asset, File *file)
 : FileBase(asset, file)
{
	reset_parameters();
	asset->format = FILE_MPEG;
	asset->byte_order = 0;
}

FileMPEG::~FileMPEG()
{
	close_file();
}

int FileMPEG::reset_parameters_derived()
{
	file = 0;
	prev_track = 0;
	data = 0;
}

int FileMPEG::open_file(int rd, int wr)
{
// Open the mpeg file
	if(wr)
	{
// Read/write or write creates video and audio streams to be multiplexed when 
// close_file is called.
		
	}
	else
	{
// Read-only determines whether it's a video, audio, or multiplexed stream
	}

	if(!(file = quicktime_open(asset->path, rd, wr)))
	{
		perror("FileMPEG::open_file");
		return 1;
	}

	if(!wr && rd) read_header();

	if(wr)
	{
		char audio_codec[5];

// Fix up the Quicktime file.
		quicktime_set_copyright(file, "Made with Broadcast 2000 for Linux");
		quicktime_set_info(file, "Quicktime for Linux");

		if(!asset->signed_ && asset->bits == BITSLINEAR8)
		{
			strcpy(audio_codec, QUICKTIME_RAW);
		}
		else
		if(asset->signed_ && asset->bits == BITSLINEAR8)
		{
			strcpy(audio_codec, QUICKTIME_TWOS);
		}
		else
		if(asset->bits == BITSULAW)
		{
			strcpy(audio_codec, QUICKTIME_ULAW);
		}
		else
		if(asset->bits == BITSIMA4)
		{
			strcpy(audio_codec, QUICKTIME_IMA4);
		} 
		else
		if(asset->bits == BITSWMX2)
		{
			strcpy(audio_codec, QUICKTIME_WMX2);
		} 
		else
		if(asset->bits == BITSLINEAR16 || asset->bits == BITSLINEAR24)
		{
			strcpy(audio_codec, QUICKTIME_TWOS);
			asset->signed_ = 1;
		}

		if(asset->video_data)
		{
			quicktime_vtracks = quicktime_set_video(file, 
						asset->layers, 
						asset->width, 
						asset->height,
						asset->frame_rate,
						asset->compression);
		}

		if(asset->audio_data)
		{
//printf("FileMov::open_file %s\n", audio_codec);
			quicktime_atracks = quicktime_set_audio(file, 
					asset->channels, 
					asset->rate, 
					asset->bits, 
					audio_codec);
		}
	}

// set the compression parameters if there are any
	quicktime_set_jpeg(file, asset->quality, 0);
	return 0;
}

int FileMPEG::create_audio_stream()
{
}

int FileMPEG::create_video_stream()
{
}


int FileMPEG::close_file_derived()
{
	if(file)
	{
	}

	if(data)
	{
		delete data;
	}
	file = 0;
	data = 0;
}

int FileMPEG::read_header()
{
}

long FileMPEG::get_audio_length()
{
	long result = quicktime_audio_length(file, 0);

	return result;
}

long FileMPEG::get_video_length()
{
	long result = quicktime_video_length(file, 0);
	return result;
}

int FileMPEG::seek_end()
{
	audio_position = get_audio_length();
	video_position = get_video_length();
	quicktime_seek_end(file);
}

int FileMPEG::seek_start()
{
	audio_position = 0;
	video_position = 0;	
	quicktime_seek_start(file);
}

long FileMPEG::get_video_position()
{
	return quicktime_video_position(file, 0);
}

long FileMPEG::get_audio_position()
{
	return quicktime_audio_position(file, 0);
}

int FileMPEG::set_audio_position(long x)
{
	audio_position = x;
// quicktime sets positions for each track seperately so store position in audio_position
	return quicktime_set_audio_position(file, x, 0);
}

int FileMPEG::set_video_position(long x)
{
	video_position = x;
	return quicktime_set_video_position(file, x, 0);
}

int FileMPEG::set_channel(int channel)
{
	this->audio_channel = channel;
}

int FileMPEG::set_layer(int layer)
{
	this->video_layer = layer;
}

int FileMPEG::write_samples(float **buffer, long len)
{
	int i;
	long bytes;
	int result = 0, track_channels = 0;

//printf("FileMPEG::write_samples 1 %x %x\n", this, FileMPEG::file);
	if(quicktime_supported_audio(file, 0))
	{
// Use Quicktime's compressor.
		result = quicktime_encode_audio(FileMPEG::file, 0, buffer, len);
	}
	else
	{
// Use our own compressor
// Originally intended to match the channels to track organization but
// Today all channels are written to track 0.
		for(int i = 0, j = 0, track = 0; i < asset->channels && result == 0; i += track_channels, track++)
		{
			track_channels = quicktime_track_channels(file, track);

			get_audio_buffer(&audio_buffer_out, len, asset->bits, track_channels);

			bytes = samples_to_raw(audio_buffer_out, 
						buffer, 
						len, 
						asset->bits, 
						track_channels,
						asset->byte_order,
						asset->signed_);      // signed_

			result = quicktime_write_audio(file, audio_buffer_out, len, track);
		}
	}
	audio_position += len;
	return result;
}

int FileMPEG::write_frames(VFrame ***buffers, long len, int use_alpha, int use_float)
{
	long bytes;
	int i, j, result = 0;

	get_video_buffer(&video_buffer_out);
	get_row_pointers(video_buffer_out, &row_pointers_out);

	for(i = 0; i < asset->layers && !result; i++)
	{
		for(j = 0; j < len; j++)
		{
			bytes = frame_to_raw(video_buffer_out, buffers[i][j], asset->width, asset->height, use_alpha, use_float, FILEBASE_RAW);
			result = quicktime_encode_video(file, row_pointers_out, i);
		}
	}

	video_position += len;
	return result;
}

int FileMPEG::read_frame(VFrame *frame, 
		int in_x1, int in_y1, int in_x2, int in_y2,
		int out_x1, int out_y1, int out_x2, int out_y2, 
		int alpha, int use_alpha, int use_float)
{
// forces read directly into frame
	read_raw(frame, in_x1, in_y1, in_x2, in_y2,
		out_x1, out_y1, out_x2, out_y2, alpha, use_alpha, use_float);

	video_position++;
	return 0;
}

VFrame* FileMPEG::read_frame(int use_alpha, int use_float)
{
// don't use alpha
// forces read into a temporary buffer
	read_raw(0, 0, 0, asset->width, asset->height, 
		0, 0, asset->width, asset->height, VMAX, 0, use_float);

	video_position++;
	return data;
}



int FileMPEG::read_raw(VFrame *frame, 
		int in_x1, int in_y1, int in_x2, int in_y2,
		int out_x1, int out_y1, int out_x2, int out_y2, 
		int alpha, int use_alpha, int use_float)
{
// want to return a pointer to a temporary buffer for drawing on a timeline
	if(!frame && !data)
	{
		frame = data = new VFrame(asset->width, asset->height);
	}

	if(!frame)
	{
// user wants frame stored in file handler's frame
		frame = data;
	}

	if(prev_frame_position == -1 || 
		prev_frame_position != video_position ||
		video_layer != prev_layer)
	{
		long i, color_channels, bytes;

		prev_layer = video_layer;
		prev_frame_position = video_position;

		get_video_buffer(&video_buffer_in);
		get_row_pointers(video_buffer_in, &row_pointers_in);

		quicktime_set_video_position(file, video_position, video_layer);
		quicktime_decode_video(file, row_pointers_in, video_layer);
		raw_to_frame(video_buffer_in, frame, 
			in_x1, in_y1, in_x2, in_y2,
			out_x1, out_y1, out_x2, out_y2, alpha, use_alpha, use_float, FILEBASE_RAW);
	}
	return 0;
}

int FileMPEG::read_samples(float *buffer, long len, int feather, 
		long lfeather_len, float lfeather_gain, float lfeather_slope, 
		long rfeather_len, float rfeather_gain, float rfeather_slope)
{
	int qt_track, qt_channel;
//printf("FileMPEG::read_samples 0\n");

	if(quicktime_supported_audio(file, 0))
	{
// Use Quicktime's decoder.
		if(feather)
		{
			get_float_buffer(&float_buffer, len);
			if(quicktime_decode_audio(file, NULL, float_buffer, len, audio_channel))
			{
				printf("FileMPEG::read_samples: quicktime_decode_audio failed\n");
				return 1;
			}

// Feather the buffer.
			overlay_float_buffer(buffer, float_buffer, len, 
				lfeather_len, lfeather_gain, lfeather_slope, 
				rfeather_len, rfeather_gain, rfeather_slope);
		}
		else
// Don't feather the buffer.
		if(quicktime_decode_audio(file, 0, buffer, len, audio_channel))
		{
			printf("FileMPEG::read_samples: quicktime_decode_audio failed\n");
			return 1;
		}
	}
	else
	{
// Use our own decoder.
//printf("FileMPEG::read_samples 1\n");
		quicktime_channel_location(file, &qt_track, &qt_channel, audio_channel);
//printf("FileMPEG::read_samples 2\n");
		get_audio_buffer(&audio_buffer_in, len, asset->bits, quicktime_track_channels(file, qt_track));
//printf("FileMPEG::read_samples 3 channels %d qt_track %d\n", quicktime_track_channels(file, qt_track), qt_track);

		if(read_raw(audio_buffer_in, len, qt_track))
		{
			printf("FileMPEG::read_samples: read_raw failed\n");
			return 1;
		}
//printf("FileMPEG::read_samples 4\n");

		raw_to_samples(buffer, 
					audio_buffer_in, 
					len, 
					asset->bits, 
					quicktime_track_channels(file, qt_track), 
					qt_channel, 
					feather, 
					lfeather_len, 
					lfeather_gain, 
					lfeather_slope,  
					rfeather_len,
					rfeather_gain, 
					rfeather_slope);
//printf("FileMPEG::read_samples 5\n");
	}

	audio_position += len;
	prev_len = len;
	return 0;
}

int FileMPEG::read_raw(char *buffer, long samples, int track)
{
	int result;

// only load if position has changed or this is the first buffer
	if(prev_buffer_position == -1 || 
		prev_buffer_position != audio_position ||
		prev_len != samples ||
		track != prev_track)
	{
		long bytes = samples * quicktime_track_channels(file, track) * (FileBase::file->bytes_per_sample(asset->bits));
		prev_buffer_position = audio_position;
		prev_len = samples;
		prev_track = track;
		result = quicktime_read_audio(file, buffer, samples, track);

		if(!result) return 1;

		if(asset->bits == BITSLINEAR8 && !asset->signed_)       // fix the unsigned numbers to signed
		{    // change unsigned to signed
			int i = 0, len_ = bytes - 4;
			while(i < len_)
			{
				buffer[i++] ^= 128;
				buffer[i++] ^= 128;
				buffer[i++] ^= 128;
				buffer[i++] ^= 128;
			}
			while(i < bytes)
			{
				buffer[i++] ^= 128;
			}
		}

//printf("FileMPEG::read_raw internal_byte_order %d asset->byte_order %d\n", internal_byte_order, asset->byte_order);
		if((!internal_byte_order && asset->byte_order) || 
			(internal_byte_order && !asset->byte_order))
		{            // swap bytes if not intel
			swap_bytes(FileBase::file->bytes_per_sample(asset->bits), (unsigned char*)buffer, bytes);
		}
	}

	return 0;
}

long FileMPEG::get_memory_usage()
{
	int result = 4096;

	if(data) result += asset->width * asset->height * sizeof(VPixel);
	if(video_buffer_in) result += asset->width * asset->height * 3;
	if(video_buffer_out) result += asset->width * asset->height * 3;
	if(audio_buffer_in) result += prev_bytes;
	if(audio_buffer_out) result += prev_bytes;
	return result;
}
