#ifndef FILEMPEG_H
#define FILEMPEG_H

#include "file.inc"
#include "filebase.h"


// Operations needed:
// 1. Open seperate audio and video streams.
// 2. Seek in seperate audio and video streams.
// 3. Read seperate streams.
// 4. Write seperate streams.
// 5. Multiplex into one stream.
// 6. Open single stream.
// 7. Seek in single stream.
// 8. Demultiplex single stream.


class FileMPEG : public FileBase
{
public:
	FileMPEG(Asset *asset, File *file);
	~FileMPEG();

	int open_file(int rd, int wr);
	int close_file_derived();
	int read_header();
	long get_video_length();
	long get_audio_length();
	long get_position();
	long get_video_position();
	long get_audio_position();
	long get_memory_usage();
	int set_video_position(long x);
	int set_audio_position(long x);
	int set_channel(int channel);  // set audio channel for reading
	int set_layer(int layer);      // set layer for reading
	int seek_end();
	int seek_start();
	int write_samples(float **buffer, long len);
	int write_frames(VFrame ***buffers, long len, int use_alpha, int use_float);
	int read_frame(VFrame *frame, 
		int in_x1, int in_y1, int in_x2, int in_y2,
		int out_x1, int out_y1, int out_x2, int out_y2, 
		int alpha, int use_alpha, int use_float);
	VFrame* read_frame(int use_alpha, int use_float);
	int read_samples(float *buffer, long len, int feather, 
		long lfeather_len, float lfeather_gain, float lfeather_slope, 
		long rfeather_len, float rfeather_gain, float rfeather_slope);

private:
// Creates an mp2 audio stream for read/write.
	int create_audio_stream();
	int create_video_stream();
	int read_raw(char *buffer, long samples, int track);  // read raw audio data
// overlay raw frame from the current channel and position
	int read_raw(VFrame *frame, 
		int in_x1, int in_y1, int in_x2, int in_y2,
		int out_x1, int out_y1, int out_x2, int out_y2, 
		int alpha, int use_alpha, int use_float);
	int reset_parameters_derived();
// current positions for when the file descriptor doesn't have the right position
	VFrame *data; // frame to return through read_frame for timeline drawing
// Temporary filenames for writing streams
	char audio_path[1024], video_path[1024];
};





#endif
