#include "assets.h"
#include "file.h"
#include "filethread.h"
#include "vframe.h"

#include <unistd.h>

FileThread::FileThread(File *file, int do_audio, int do_video) : Thread()
{
	this->file = file;
	this->do_audio = do_audio;
	this->do_video = do_video;
	for(int i = 0; i < TOTAL_BUFFERS; i++)
	{
		output_lock[i].lock();
		input_lock[i].unlock();
	}

	current_buffer = 0;
	return_value = 0;
	local_buffer = 0;
}

FileThread::~FileThread()
{
}

void FileThread::run()
{
	int done = 0;

	while(!done)
	{
		output_lock[local_buffer].lock();

		if(!last_buffer[local_buffer])
		{
			if(output_size[local_buffer])
			{
				if(do_audio)
				{
					return_value = file->write_samples(audio_buffer[local_buffer], output_size[local_buffer]);
				}
				else
				if(do_video)
				{
					return_value = file->write_frames(video_buffer[local_buffer], output_size[local_buffer], use_alpha, use_float);
				}
			}
			else
				return_value = 0;
		}
		else
		done = 1;

		input_lock[local_buffer].unlock();
		local_buffer++;
		if(local_buffer >= total_buffers) local_buffer = 0;
	}
}



int FileThread::stop_writing()
{
// signal thread to stop
	input_lock[current_buffer].lock();

	last_buffer[current_buffer] = 1;
	output_lock[current_buffer].unlock();

	swap_buffer();

// wait for thread to finish
	join();

// delete buffers
	int i;
	for(int buffer = 0; buffer < total_buffers; buffer++)
	{
		if(do_audio)
		{
			for(int channel = 0; channel < file->asset->channels; channel++)
			{
				delete audio_buffer[buffer][channel];
			}
			delete audio_buffer[buffer];
		}
		
		if(do_video)
		{
			
			for(int layer = 0; layer < file->asset->layers; layer++)
			{
				for(i = 0; i < buffer_size; i++)
				{
					delete video_buffer[buffer][layer][i];
				}

				delete video_buffer[buffer][layer];
			}

			delete video_buffer[buffer];
		}
	}
}

int FileThread::start_writing(long buffer_size, int use_alpha, int total_buffers)
{
// allocate buffers
	int i;
	this->buffer_size = buffer_size;
	this->total_buffers = total_buffers;

	for(int buffer = 0; buffer < total_buffers; buffer++)
	{
		if(do_audio)
		{
			audio_buffer[buffer] = new float*[file->asset->channels];

			for(int channel = 0; channel < file->asset->channels; channel++)
			{
				audio_buffer[buffer][channel] = new float[buffer_size];
			}
		}
		if(do_video)
		{
			video_buffer[buffer] = new VFrame**[file->asset->layers];

			for(int layer = 0; layer < file->asset->layers; layer++)
			{
				video_buffer[buffer][layer] = new VFrame*[buffer_size];
				for(i = 0; i < buffer_size; i++)
				{
					video_buffer[buffer][layer][i] = new VFrame(file->asset->width, file->asset->height, use_alpha);
					video_buffer[buffer][layer][i]->clear_frame();
				}
			}
		}
	}

	for(int i = 0; i < total_buffers; i++)
	{
		last_buffer[i] = 0;
	}

	synchronous = 1;
	
	start();
}

float** FileThread::get_audio_buffer()
{
	input_lock[current_buffer].lock();

	return audio_buffer[current_buffer];
}

VFrame*** FileThread::get_video_buffer()
{
	input_lock[current_buffer].lock();

	return video_buffer[current_buffer];
}

int FileThread::write_buffer(long size, int use_alpha, int use_float, int compressed)
{
	this->use_alpha = use_alpha;
	this->use_float = use_float;
	output_size[current_buffer] = size;
	is_compressed[current_buffer] = compressed;

// unlock the output lock
	output_lock[current_buffer].unlock();

	swap_buffer();

	return return_value;
}

int FileThread::swap_buffer()
{
	current_buffer++;
	if(current_buffer >= total_buffers) current_buffer = 0;
}
