#include "playbackprefs.h"
#include "preferences.h"

PlaybackPrefs::PlaybackPrefs(PreferencesWindow *pwindow)
 : BC_SubWindow(150, 10, PREFSUBWINDOW_W, PREFSUBWINDOW_H, MEGREY)
{ this->pwindow = pwindow; }

int PlaybackPrefs::create_objects()
{
	int x = 10, y = 10;
	char string[1024];

	add_border(get_resources()->get_bg_shadow1(),
		get_resources()->get_bg_shadow2(),
		get_resources()->get_bg_color(),
		get_resources()->get_bg_light2(),
		get_resources()->get_bg_light1());

	add_tool(new BC_Title(x, y, "Playback", LARGEFONT, RED));
	y += 35;
	add_tool(new BC_Title(x, y, "Device", MEDIUMFONT, BLACK));
	x += 115;
	add_tool(new BC_Title(x, y, "Bits", MEDIUMFONT, BLACK));
	x += 65;
	add_tool(new BC_Title(x, y, "Channels", MEDIUMFONT, BLACK));
	y += 20;
	x = 10;
	add_tool(outpath = new PlaybackOutPath(x, y, pwindow, pwindow->playback_device));
	x += 115;
	sprintf(string, "%d", pwindow->playback_bitrate);
	add_tool(outbits = new PlaybackOutBits(x, y, pwindow, string));
	x += 65;
	sprintf(string, "%d", pwindow->playback_channels);
	add_tool(outchannels = new PlaybackOutChannels(x, y, pwindow, string));
	y += 35;

	x = 10;
	add_tool(new BC_Title(x, y, "Samples to read from disk at a time:", MEDIUMFONT, BLACK));
	sprintf(string, "%d", pwindow->playback_read_length);
	add_tool(read_length = new PlaybackReadLength(x + 275, y, pwindow, this, string));
	y += 30;
	add_tool(new BC_Title(x, y, "Samples to send to console at a time:", MEDIUMFONT, BLACK));
	sprintf(string, "%d", pwindow->playback_module_fragment);
	add_tool(module_fragment = new PlaybackModuleFragment(x + 275, y, pwindow, this, string));
	y += 30;
	add_tool(new BC_Title(x, y, "Preload buffer for multiplexed streams:", MEDIUMFONT, BLACK));
	sprintf(string, "%d", pwindow->playback_preload);
	add_tool(playback_preload = new PlaybackPreload(x + 275, y, pwindow, this, string));
	y += 35;
	
	add_tool(disable_noedits = new PlaybackDisableNoEdits(pwindow, pwindow->test_playback_edits, y));
	y += 20;
	add_tool(view_follows = new PlaybackViewFollows(pwindow, pwindow->view_follows_playback, y));
	y += 20;
	add_tool(software_timer = new PlaybackSoftwareTimer(pwindow, pwindow->playback_software_timer, y));
	y += 20;
	add_tool(real_time = new PlaybackRealTime(pwindow, pwindow->real_time_playback, y));
}

PlaybackPrefs::~PlaybackPrefs()
{
	delete outpath;
	delete outbits;
	delete outchannels;
	//delete buffer_size;
	//delete buffer_bytes;
	delete view_follows;
	delete software_timer;
	delete real_time;
	delete module_fragment;
	delete read_length;
	delete disable_noedits;
	delete playback_preload;
}


int PlaybackPrefs::get_buffer_bytes()
{
	return pwindow->playback_bitrate / 8 * pwindow->playback_channels * pwindow->playback_buffer;
}








PlaybackOutPath::PlaybackOutPath(int x, int y, PreferencesWindow *pwindow, char *text)
 : BC_TextBox(x, y, 100, text)
{ this->pwindow = pwindow; }

int PlaybackOutPath::handle_event() { strcpy(pwindow->playback_device, get_text()); }

PlaybackOutBits::PlaybackOutBits(int x, int y, PreferencesWindow *pwindow, char *text)
 : BC_TextBox(x, y, 50, text)
{ this->pwindow = pwindow; }

int PlaybackOutBits::handle_event() { pwindow->playback_bitrate = atol(get_text()); }

PlaybackOutChannels::PlaybackOutChannels(int x, int y, PreferencesWindow *pwindow, char *text)
 : BC_TextBox(x, y, 50, text)
{ this->pwindow = pwindow; }

int PlaybackOutChannels::handle_event() { pwindow->playback_channels = atol(get_text()); }




PlaybackReadLength::PlaybackReadLength(int x, int y, PreferencesWindow *pwindow, PlaybackPrefs *playback, char *text)
 : BC_TextBox(x, y, 100, text)
{ this->pwindow = pwindow; this->playback = playback; }

int PlaybackReadLength::handle_event() 
{ 
	pwindow->playback_read_length = atol(get_text()); 
}




PlaybackModuleFragment::PlaybackModuleFragment(int x, int y, PreferencesWindow *pwindow, PlaybackPrefs *playback, char *text)
 : BC_TextBox(x, y, 100, text)
{ this->pwindow = pwindow; }

int PlaybackModuleFragment::handle_event() 
{ 
	pwindow->playback_module_fragment = atol(get_text()); 
}


PlaybackPreload::PlaybackPreload(int x, int y, PreferencesWindow *pwindow, PlaybackPrefs *playback, char *text)
 : BC_TextBox(x, y, 100, text)
{ this->pwindow = pwindow; this->playback = playback; }

int PlaybackPreload::handle_event() 
{ 
	pwindow->playback_preload = atol(get_text()); 
}





PlaybackBufferSize::PlaybackBufferSize(int x, int y, PreferencesWindow *pwindow, PlaybackPrefs *playback, char *text)
 : BC_TextBox(x, y, 100, text)
{ this->pwindow = pwindow; this->playback = playback; }

int PlaybackBufferSize::handle_event() 
{ 
	pwindow->playback_buffer = atol(get_text()); 

	//playback->buffer_bytes->update_bytes();
}

PlaybackBufferBytes::PlaybackBufferBytes(int x, int y, PreferencesWindow *pwindow, PlaybackPrefs *playback, char *text)
 : BC_Title(x, y, text)
{
	this->pwindow = pwindow; this->playback = playback;
}
int PlaybackBufferBytes::update_bytes()
{
	sprintf(string, "%d", playback->get_buffer_bytes());
	update(string);
}

PlaybackDisableNoEdits::PlaybackDisableNoEdits(PreferencesWindow *pwindow, int value, int y)
 : BC_CheckBox(10, y, 16, 16, value, "Disable tracks when no edits.")
{ this->pwindow = pwindow; }

int PlaybackDisableNoEdits::handle_event() { pwindow->test_playback_edits = get_value(); }




PlaybackViewFollows::PlaybackViewFollows(PreferencesWindow *pwindow, int value, int y)
 : BC_CheckBox(10, y, 16, 16, value, "View follows playback")
{ this->pwindow = pwindow; }

int PlaybackViewFollows::handle_event() { pwindow->view_follows_playback = get_value(); }




PlaybackSoftwareTimer::PlaybackSoftwareTimer(PreferencesWindow *pwindow, int value, int y)
 : BC_CheckBox(10, y, 16, 16, value, "Use software for positioning information")
{ this->pwindow = pwindow; }

int PlaybackSoftwareTimer::handle_event() { pwindow->playback_software_timer = get_value(); }




PlaybackRealTime::PlaybackRealTime(PreferencesWindow *pwindow, int value, int y)
 : BC_CheckBox(10, y, 16, 16, value, "Audio playback in real time priority (root only)")
{ this->pwindow = pwindow; }

int PlaybackRealTime::handle_event() { pwindow->real_time_playback = get_value(); }
