#ifndef RECORDTHREAD_H
#define RECORDTHREAD_H

#include "audiodevice.inc"
#include "bcbase.h"
#include "file.inc"
#include "mutex.h"
#include "record.inc"
#include "recordgui.inc"
#include "recordengine.inc"
#include "thread.h"


class RecordAudio : public Thread
{
public:
	RecordAudio(Record *record, 
				RecordEngine *engine, 
				AudioDevice *device,
				int write_file);
	~RecordAudio();
	
	void run();
	int start_recording();            // start saving audio data to file
	int stop_recording();
	int pause_recording();
	int resume_recording();
	int wait_for_completion();

	int set_parameters(File *file,
					RecordGUI *gui,
					BC_Meter **meter, 
					long fragment_size, 
					long buffer_size, 
					int record_channels, 
					long *dc_offset, 
					int realtime);
	long get_position();
	int set_position(long position);     // seek to a new location in the file 
	long absolute_position();

	int write_buffer();           // write the buffer
	int set_dc_offset(long new_offset, int number);
	long get_dc_offset(long *dc_offset, RecordGUIDCOffsetText **dc_offset_text);




	Mutex completed;
	Mutex monitor_paused;
	Mutex pause_lock;
	Mutex timer_lock;
	int is_recording;
	int is_paused;
	File *file;
	RecordEngine *engine;
	AudioDevice *device;
	BC_Meter **meter;
	Timer timer;
	long total_samples;
	float **input;
	long *dc_offset;
	long buffer_size, fragment_size, fragment_position;
	int record_channels;
	int write_file;
	Record *record;
	RecordGUI *gui;
	int quit_when_completed;
};



#endif
