#include "preferences.h"
#include "recordprefs.h"

RecordPrefs::RecordPrefs(PreferencesWindow *pwindow)
 : BC_SubWindow(150, 10, PREFSUBWINDOW_W, PREFSUBWINDOW_H, MEGREY)
{ this->pwindow = pwindow; }

int RecordPrefs::create_objects()
{
	add_border(get_resources()->get_bg_shadow1(),
		get_resources()->get_bg_shadow2(),
		get_resources()->get_bg_color(),
		get_resources()->get_bg_light2(),
		get_resources()->get_bg_light1());
	add_tool(new BC_Title(10, 5, "Recording", LARGEFONT, RED));
	add_tool(new BC_Title(10, 35, "Record Device"));
	add_tool(new BC_Title(125, 35, "Bits"));
	add_tool(new BC_Title(205, 35, "Channels"));
	add_tool(new BC_Title(10, 95, "Duplex Device"));
	add_tool(new BC_Title(125, 95, "Bits"));
	add_tool(new BC_Title(205, 95, "Channels"));
	add_tool(new BC_Title(115, 155, "Samples to write to disk at a time"));

	add_tool(inpath = new RecordInPath(pwindow, pwindow->record_device));
	char string[1024];
	sprintf(string, "%d", pwindow->record_bitrate);
	add_tool(inbits = new RecordInBits(pwindow, string));
	sprintf(string, "%d", pwindow->record_channels);
	add_tool(inchannels = new RecordInChannels(pwindow, string));
	sprintf(string, "%ld", pwindow->record_write_length);
	add_tool(write_length = new RecordWriteLength(pwindow, string));
	add_tool(duplexenable = new DuplexEnable(pwindow, pwindow->enable_duplex));
	add_tool(realtime = new RecordRealTime(pwindow, pwindow->real_time_record));

	add_tool(outpath = new DuplexOutPath(pwindow, pwindow->duplex_device));
	sprintf(string, "%d", pwindow->duplex_bitrate);
	add_tool(outbits = new DuplexOutBits(pwindow, string));
	sprintf(string, "%d", pwindow->duplex_channels);
	add_tool(outchannels = new DuplexOutChannels(pwindow, string));
}

RecordPrefs::~RecordPrefs()
{
	delete inpath;
	delete inbits;
	delete inchannels;
	delete outpath;
	delete outbits;
	delete outchannels;
	delete write_length;
	delete duplexenable;
	delete realtime;
}




RecordInPath::RecordInPath(PreferencesWindow *pwindow, char *text)
 : BC_TextBox(10, 60, 100, text)
{ this->pwindow = pwindow; }

int RecordInPath::handle_event()
{ strcpy(pwindow->record_device, get_text()); }

RecordInBits::RecordInBits(PreferencesWindow *pwindow, char *text)
 : BC_TextBox(125, 60, 50, text)
{ this->pwindow = pwindow; }

int RecordInBits::handle_event()
{ pwindow->record_bitrate = atol(get_text()); }

RecordInChannels::RecordInChannels(PreferencesWindow *pwindow, char *text)
 : BC_TextBox(205, 60, 50, text)
{ this->pwindow = pwindow; }

int RecordInChannels::handle_event()
{ pwindow->record_channels = atol(get_text()); }




DuplexOutPath::DuplexOutPath(PreferencesWindow *pwindow, char *text)
 : BC_TextBox(10, 120, 100, text)
{ this->pwindow = pwindow; }

int DuplexOutPath::handle_event()
{ strcpy(pwindow->duplex_device, get_text()); }

DuplexOutBits::DuplexOutBits(PreferencesWindow *pwindow, char *text)
 : BC_TextBox(125, 120, 50, text)
{ this->pwindow = pwindow; }

int DuplexOutBits::handle_event()
{ pwindow->duplex_bitrate = atol(get_text()); }

DuplexOutChannels::DuplexOutChannels(PreferencesWindow *pwindow, char *text)
 : BC_TextBox(205, 120, 50, text)
{ this->pwindow = pwindow; }

int DuplexOutChannels::handle_event()
{ pwindow->duplex_channels = atol(get_text()); }



RecordWriteLength::RecordWriteLength(PreferencesWindow *pwindow, char *text)
 : BC_TextBox(10, 155, 100, text)
{ this->pwindow = pwindow; }

int RecordWriteLength::handle_event()
{ pwindow->record_write_length = atol(get_text()); }




RecordRealTime::RecordRealTime(PreferencesWindow *pwindow, int value)
 : BC_CheckBox(10, 190, 16, 16, value, "Record in realtime priority (root only)")
{ this->pwindow = pwindow; }

int RecordRealTime::handle_event()
{
	pwindow->real_time_record = get_value();
}

DuplexEnable::DuplexEnable(PreferencesWindow *pwindow, int value)
 : BC_CheckBox(10, 215, 16, 16, value, "Enable full duplex")
{ this->pwindow = pwindow; }

int DuplexEnable::handle_event()
{
	pwindow->enable_duplex = get_value();
}
