#ifndef RECORDVIDEO_H
#define RECORDVIDEO_H

#include "file.inc"
#include "mutex.h"
#include "record.inc"
#include "recordgui.inc"
#include "thread.h"
#include "videodevice.inc"

// Default behavior is to read frames and flash to display.
// Optionally writes to disk.

class RecordVideo : public Thread
{
public:
	RecordVideo(Record *record, 
		RecordEngine *engine, 
		VideoDevice *vdevice, 
		int write_file);
	~RecordVideo();

	void run();
	int start_recording();
	int stop_recording();
	int pause_recording();
	int resume_recording();
	int wait_for_completion();     // For recording to a file.
	int set_parameters(File *file,
							RecordGUI *gui,
							int buffer_size,    // number of frames to write to disk at a time
							int realtime);
	int unhang_thread();

	VideoDevice *vdevice;
	Record *record;
	RecordEngine *engine;
	int write_file;  // Set if writing to disk.
	Mutex pause_lock;    // Wait during pause commands until it's really paused.
	Mutex monitor_paused;  // For pausing
	Mutex completed; // For recording
	RecordGUI *gui;
	int is_recording;
	int is_paused;
	File *file;
	long total_frames;  // Number of frames actually captured
	int buffer_size;    // number of frames to write to disk at a time
	Mutex unhang_lock;
	long buffer_position;   // Position in output buffer being captured to
	int quit_when_completed;
};


#endif
