#ifndef BCBITMAP_H
#define BCBITMAP_H

#include <X11/Xlib.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>

#include "bctool.inc"
#include "bcwindow.inc"
#include "sizes.h"
#include "vframe.inc"

class BC_Bitmap
{
public:
	BC_Bitmap(BC_Window *top_level, int w, int h, int depth);
	virtual ~BC_Bitmap();

// transfer VFrame with alpha
	int read_frame(VFrame *frame, int x1, int y1, int x2, int y2, int use_alpha);
// transfer raw rgb bitmap data with no alpha to ximage
	int read_frame(unsigned char *frame, int in_w, int in_h);


// If dont_wait is true, the XSync call comes  before the flash
	int write_drawable(Drawable &pixmap, int dest_x, int dest_y, int source_x, int source_y, int w, int h, int dont_wait);
// the bitmap must be wholly contained in the source during a GetImage
	int read_drawable(Drawable &pixmap, int source_x, int source_y);

	int rotate_90(int side);
	int get_w();
	int get_h();

	int w, h, depth;
	unsigned char **row_data;

private:
	int transfer_direct(unsigned char *frame, int in_w, int in_h);
	int transfer_scale(unsigned char *frame, int in_w, int in_h);
	int transfer_direct(VFrame *frame, int x1, int y1, int x2, int y2, int use_alpha);
	int transfer_scale(VFrame *frame, int x1, int y1, int x2, int y2, int use_alpha);
	int transfer_row_raw_direct(unsigned char *output, unsigned char *input, int width);
	int transfer_row_raw_scale(unsigned char *output, unsigned char *input, int *column_table);
	int transfer_row_direct_8(unsigned char *output, VPixel *input, int width, int use_alpha);
	int transfer_row_scale_8(unsigned char *output, VPixel *input, int *column_table, int use_alpha);
	int transfer_row_direct_16(unsigned TWO *output, VPixel *input, int width, int use_alpha);
	int transfer_row_scale_16(unsigned TWO *output, VPixel *input, int *column_table, int use_alpha);
	int transfer_row_direct_24(unsigned char *output, VPixel *input, int width, int use_alpha);
	int transfer_row_scale_24(unsigned char *output, VPixel *input, int *column_table, int use_alpha);
	int transfer_row_direct_32(unsigned char *output, VPixel *input, int width, int use_alpha);
	int transfer_row_scale_32(unsigned char *output, VPixel *input, int *column_table, int use_alpha);
	int transfer_pixel_8(unsigned char *result, VPixel *input);
	int transfer_raw_pixel_8(unsigned char *result, unsigned char *input);
	int transfer_pixel_8_alpha(unsigned char *result, VPixel *input);
	int transfer_pixel_16(unsigned TWO *result, VPixel *input);
	int transfer_raw_pixel_16(unsigned TWO *result, unsigned char *input);
	int transfer_pixel_16_alpha(unsigned TWO *result, VPixel *input);
	int transfer_pixel_24(unsigned char **result, VPixel *input);
	int transfer_pixel_24_alpha(unsigned char **result, VPixel *input);
	int transfer_pixel_32(unsigned char **result, VPixel *input);
	int transfer_raw_pixel_24(unsigned char *result, unsigned char *input);
	int allocate_data();
	int delete_data();

	unsigned char *data;
	XImage *ximage;
	XShmSegmentInfo shm_info;
	int bits_per_pixel;
	XEvent shm_event;
	long shm_event_type;
	int client_byte_order, server_byte_order;
	BC_Window *top_level;
};

#endif
