#ifndef VCONFIG_H
#define VCONFIG_H

#include "vframe.inc"

class VFrame
{
public:
	VFrame(VPixel *data, int w, int h, int alpha = 1);  // For shared with alpha
	VFrame(unsigned char *data, int w, int h);  // For shared without alpha
	VFrame(int w, int h, int alpha = 1);  // For privately owned data in new frame
	VFrame();  // For compressed data
	~VFrame();

// Get a compressed data buffer
	int allocate_compressed_data(long bytes);
// Get the compressed data pointer
	char* get_compressed_data();
	long get_compressed_allocated();
	long get_compressed_size();
	long set_compressed_size(long size);

// Reallocate uncompressed buffer with or without alpha
	int reallocate(VPixel *data, int w, int h, int alpha);

// if frame points to the same data as this return 1
	int equals(VFrame *frame);

// direct copy with no alpha
	int copy_from(VFrame *frame);
// direct copy with alpha
	int replace_from(VFrame *frame, int alpha, int use_alpha, int use_float);
	int get_scale_tables(int *column_table, int *row_table, 
			int in_x1, int in_y1, int in_x2, int in_y2,
			int out_x1, int out_y1, int out_x2, int out_y2);

	int apply_fade(int factor);
	int clear_frame();
	VPixel** get_rows();        // return an array of pointers to rows
	unsigned char* get_rows_noalpha();  // return a block
	int get_w();
	int get_h();
	int is_alpha();
	int w, h;

private:
	int allocate_uncompressed_data(VPixel *data, int w, int h, int alpha);
	int clear_objects();
	int reset_flags();
	int clear_pixel(VPixel *pixel);
	int shared; // data is shared with something else
	int has_alpha; // data is either RGBA or RGB
	VPixel **rows;
// Word size defaults to unsigned char when no alpha.
	unsigned char *rows_noalpha; // With no alpha, the rows are sequential in one block
	char *compressed_data;     // With compressed data.
	long compressed_allocated;  // Allocated space
	long compressed_size;   // Size of current image
};


#endif
