#include "filehtal.h"
#include "mpeg.h"
#include "overlayframe.h"
#include "vframe.h"

main(int argc, char *argv[])
{
	MpegMain *plugin;

	plugin = new MpegMain(argc, argv);
	plugin->plugin_run();
	delete plugin;
}

MpegMain::MpegMain(int argc, char *argv[])
 : PluginIOClient(argc, argv)
{
	reset_parameters();
}

MpegMain::~MpegMain()
{
	close_file();
}

int MpegMain::reset_parameters()
{
	overlayer = 0;
	rgb24_frame = 0;
	temp_frame = 0;
	temp_frame_buffer = 0;
	audio_allocated = 0;
	audio_buffer = 0;
	current_channel = 0;
	defaults = 0;
	mpeg = 0;
}

char* MpegMain::plugin_title() { return "MPEG"; }
int MpegMain::plugin_is_audio() { return 1; }
int MpegMain::plugin_is_video() { return 1; }

int MpegMain::get_audio_parameters()
{
	return 0;
}

int MpegMain::get_video_parameters()
{
	return 0;
}

int MpegMain::load_defaults()
{
	return 0;
}

int MpegMain::save_defaults()
{
	return 0;
}

int MpegMain::check_header(char *path)
{
	int result = 0;
	result = mpeg2_check_sig(path);
	return result;
}

int MpegMain::open_file(char *path, int rd, int wr)
{
	mpeg = mpeg2_open(path);
	if(!mpeg) return 1;
	read_header();
	return 0;
}


int MpegMain::close_file()
{
	if(defaults) delete defaults;
	if(temp_frame)
	{
		delete temp_frame;
		delete temp_frame_buffer;
	}
	if(rgb24_frame)
	{
		delete rgb24_frame;
	}
	if(overlayer)
	{
		delete overlayer;
	}
	if(mpeg) mpeg2_close(mpeg);
	if(audio_buffer) delete audio_buffer;
	reset_parameters();
	return 0;
}

int MpegMain::read_header()
{
	has_audio = mpeg2_has_audio(mpeg);
	if(has_audio)
	{
		sample_rate = mpeg2_sample_rate(mpeg);
		channels = mpeg2_audio_channels(mpeg);
	}
	has_video = mpeg2_has_video(mpeg);
	if(has_video)
	{
		layers = 1;
		frame_rate = mpeg2_frame_rate(mpeg);
		width = mpeg2_video_width(mpeg);
		height = mpeg2_video_height(mpeg);
	}
	return 0;
}

long MpegMain::get_alength()
{
	return mpeg2_audio_samples(mpeg);
}

long MpegMain::get_vlength()
{
	return mpeg2_video_frames(mpeg);
}

int MpegMain::seek_end()
{
	mpeg2_set_sample(mpeg, get_alength());
	mpeg2_set_frame(mpeg, get_vlength());
}

int MpegMain::seek_start()
{
	mpeg2_set_sample(mpeg, 0);
	mpeg2_set_frame(mpeg, 0);
}

long MpegMain::get_aposition()
{
	return mpeg2_get_sample(mpeg);
}

long MpegMain::get_vposition()
{
	return mpeg2_get_frame(mpeg);
}

int MpegMain::set_aposition(long sample)
{
	return mpeg2_set_sample(mpeg, sample);
}

int MpegMain::set_vposition(long frame)
{
	return mpeg2_set_frame(mpeg, frame);
}

int MpegMain::set_channel(int channel)
{
	current_channel = channel;
	return 0;
}

int MpegMain::set_layer(int layer)
{
// Only 1 layer supported
	return 0;
}

int MpegMain::read_samples(float *buffer, long len, 
	int feather, 
	long lfeather_len, float lfeather_gain, float lfeather_slope, 
	long rfeather_len, float rfeather_gain, float rfeather_slope)
{
	int result;

// Allocate a new buffer
	if(audio_allocated < len)
	{
		delete audio_buffer;
		audio_buffer = 0;
	}
	if(!audio_buffer)
	{
		audio_buffer = new float[len];
	}

// Overlay from the buffer
	if(feather)
	{
		long rfeather_start = (long)(len - rfeather_len);        // start of right feather relative to input buffer
		float feather_current;     // input position for right feather
		float current_gain;
		long output_current = 0;  // position in output buffer
		float *audio_ptr = audio_buffer;

		result = mpeg2_read_audio(mpeg, audio_buffer, 0, current_channel, len);
		for(feather_current = 0; feather_current < lfeather_len; 
			output_current++, feather_current++)
		{
			current_gain = lfeather_gain + lfeather_slope * feather_current;
			buffer[output_current] += *audio_ptr++ * current_gain;
		}

		for( ; output_current < rfeather_start; output_current++)
		{
			buffer[output_current] += *audio_ptr++;
		}

		for(feather_current = 0; feather_current < rfeather_len; 
			output_current++, feather_current++)
		{
			current_gain = rfeather_gain + rfeather_slope * feather_current;
			buffer[output_current] += *audio_ptr++ * current_gain;
		}
	}
	else
	{
		result = mpeg2_read_audio(mpeg, buffer, 0, current_channel, len);
	}
	return result;
}

int MpegMain::read_frame_ptr(int use_alpha, 
		int use_float, 
		long &shared_id, 
		long &shared_size)
{
	int result = 0;
	if(!temp_frame)
	{
		temp_frame_buffer = new PluginBuffer(width * height, sizeof(VPixel));
		temp_frame = new VFrame((VPixel*)temp_frame_buffer->get_data(), width, height, 1);
	}

	result = read_rgb24_frame(temp_frame, 
			0, 0, width, height,
			0, 0, width, height,
			VMAX, 0, use_float, 0);

	shared_id = temp_frame_buffer->get_id();
	shared_size = width * height;
	return result;
}

int MpegMain::read_frame(VFrame *frame, 
	float in_x1, float in_y1, float in_x2, float in_y2,
	float out_x1, float out_y1, float out_x2, float out_y2, 
	int alpha, int use_alpha, int use_float, int interpolate)
{
	int result = read_rgb24_frame(frame, 
			in_x1, in_y1, in_x2, in_y2,
			out_x1, out_y1, out_x2, out_y2, 
			alpha, use_alpha, use_float, interpolate);

	return result;
}

int MpegMain::read_raw_frame_possible()
{
	return 1;
}

int MpegMain::read_raw_frame(VFrame *frame)
{
	int result = mpeg2_read_frame(mpeg, frame->get_rows_noalpha());
	return result;
}

int MpegMain::read_rgb24_frame(VFrame *frame, 
		float in_x1, float in_y1, float in_x2, float in_y2,
		float out_x1, float out_y1, float out_x2, float out_y2, 
		int alpha, int use_alpha, int use_float, int interpolate)
{
	int result = 0;

	if(!rgb24_frame)
	{
		rgb24_frame = new unsigned char[width * height * 3];
	}

	result = mpeg2_read_frame(mpeg, rgb24_frame);

	if(overlayer)
	{
		if(!overlayer->compare_with(use_float, use_alpha, interpolate, NORMAL))
		{
			delete overlayer;
			overlayer = 0;
		}
	}

	if(!overlayer)
	{
		overlayer = new OverlayFrame(use_alpha, use_float, interpolate, NORMAL, OVERLAY_RGB);
	}

	overlayer->overlay(frame, rgb24_frame,
		in_x1, in_y1, in_x2, in_y2,
		out_x1, out_y1, out_x2, out_y2, 
		alpha, width, height);

	return result;
}
