#ifndef AC3_H
#define AC3_H

#define MPEG2_FLOAT32 float
#define MPEG2_INT32   int

extern int mpeg2_ac3_samplerates[3];

/* Exponent strategy constants */
#define MPEG2_EXP_REUSE (0)
#define MPEG2_EXP_D15   (1)
#define MPEG2_EXP_D25   (2)
#define MPEG2_EXP_D45   (3)

/* Delta bit allocation constants */
#define DELTA_BIT_REUSE (0)
#define DELTA_BIT_NEW (1)
#define DELTA_BIT_NONE (2)
#define DELTA_BIT_RESERVED (3)

typedef struct
{
	MPEG2_FLOAT32 fbw[5][256];
	MPEG2_FLOAT32 lfe[256];
} mpeg2_stream_coeffs_t;

typedef struct
{
	unsigned long magic;
/* Bit stream identification == 0x8 */
	unsigned int bsid;	
/* Bit stream mode */
	unsigned int bsmod;
/* Audio coding mode */
	unsigned int acmod;
/* If we're using the centre channel then */
/* centre mix level */
		unsigned int cmixlev;
/* If we're using the surround channel then */
/* surround mix level */
		unsigned int surmixlev;
/* If we're in 2/0 mode then */
/* Dolby surround mix level - NOT USED - */
		unsigned int dsurmod;
/* Low frequency effects on */
	unsigned int lfeon;
/* Dialogue Normalization level */
	unsigned int dialnorm;
/* Compression exists */
	unsigned int compre;
/* Compression level */
		unsigned int compr;
/* Language code exists */
	unsigned int langcode;
/* Language code */
		unsigned int langcod;
/* Audio production info exists*/
	unsigned int audprodie;
		unsigned int mixlevel;
		unsigned int roomtyp;
/* If we're in dual mono mode (acmod == 0) then extra stuff */
		unsigned int dialnorm2;
		unsigned int compr2e;
			unsigned int compr2;
		unsigned int langcod2e;
			unsigned int langcod2;
		unsigned int audprodi2e;
			unsigned int mixlevel2;
			unsigned int roomtyp2;
/* Copyright bit */
	unsigned int copyrightb;
/* Original bit */
	unsigned int origbs;
/* Timecode 1 exists */
	unsigned int timecod1e;
/* Timecode 1 */
		unsigned int timecod1;
/* Timecode 2 exists */
	unsigned int timecod2e;
/* Timecode 2 */
		unsigned int timecod2;
/* Additional bit stream info exists */
	unsigned int addbsie;
/* Additional bit stream length - 1 (in bytes) */
		unsigned int addbsil;
/* Additional bit stream information (max 64 bytes) */
		unsigned char addbsi[64];

/* Information not in the AC-3 bitstream, but derived */
/* Number of channels (excluding LFE)
 * Derived from acmod */
	unsigned int nfchans;
} mpeg2_ac3bsi_t;

typedef struct 
{
	unsigned long magic1;
/* block switch bit indexed by channel num */
	unsigned short blksw[5];
/* dither enable bit indexed by channel num */
	unsigned short dithflag[5];
/* dynamic range gain exists */
	unsigned int dynrnge;
/* dynamic range gain */
		unsigned int dynrng;
/* if acmod==0 then */
/* dynamic range 2 gain exists */
	unsigned int dynrng2e;
/* dynamic range 2 gain */
		unsigned int dynrng2;
/* coupling strategy exists */
	unsigned int cplstre;
/* coupling in use */
		unsigned int cplinu;
/* channel coupled */
			unsigned short chincpl[5];
/* if acmod==2 then */
/* Phase flags in use */
				unsigned int phsflginu;
/* coupling begin frequency code */
			unsigned int cplbegf;
/* coupling end frequency code */
			unsigned int cplendf;
/* coupling band structure bits */
			unsigned short cplbndstrc[18];
/* Do coupling co-ords exist for this channel? */
			unsigned short cplcoe[5];
/* Master coupling co-ordinate */
			unsigned short mstrcplco[5];
/* Per coupling band coupling co-ordinates */
			unsigned short cplcoexp[5][18];
			unsigned short cplcomant[5][18];
/* Phase flags for dual mono */
			unsigned short phsflg[18];
/* Is there a rematrixing strategy */
	unsigned int rematstr;
/* Rematrixing bits */
		unsigned short rematflg[4];
/* Coupling exponent strategy */
	unsigned int cplexpstr;
/* Exponent strategy for full bandwidth channels */
	unsigned short chexpstr[5];
/* Exponent strategy for lfe channel */
	unsigned int lfeexpstr;
/* Channel bandwidth for independent channels */
	unsigned short chbwcod[5];
/* The absolute coupling exponent */
		unsigned int cplabsexp;
/* Coupling channel exponents (D15 mode gives 18 * 12 /3  encoded exponents */
		unsigned short cplexps[18 * 12 / 3];
/* Sanity checking constant */
	unsigned long magic2;
/* fbw channel exponents */
	unsigned short exps[5][252 / 3];
/* channel gain range */
	unsigned short gainrng[5];
/* low frequency exponents */
	unsigned short lfeexps[3];

/* Bit allocation info */
	unsigned int baie;
/* Slow decay code */
		unsigned int sdcycod;
/* Fast decay code */
		unsigned int fdcycod;
/* Slow gain code */
		unsigned int sgaincod;
/* dB per bit code */
		unsigned int dbpbcod;
/* masking floor code */
		unsigned int floorcod;

/* SNR offset info */
	unsigned int snroffste;
/* coarse SNR offset */
		unsigned int csnroffst;
/* coupling fine SNR offset */
		unsigned int cplfsnroffst;
/* coupling fast gain code */
		unsigned int cplfgaincod;
/* fbw fine SNR offset */
		unsigned short fsnroffst[5];
/* fbw fast gain code */
		unsigned short fgaincod[5];
/* lfe fine SNR offset */
		unsigned int lfefsnroffst;
/* lfe fast gain code */
		unsigned int lfefgaincod;
	
/* Coupling leak info */
	unsigned int cplleake;
/* coupling fast leak initialization */
		unsigned int cplfleak;
/* coupling slow leak initialization */
		unsigned int cplsleak;
	
/* delta bit allocation info */
	unsigned int deltbaie;
/* coupling delta bit allocation exists */
		unsigned int cpldeltbae;
/* fbw delta bit allocation exists */
		unsigned short deltbae[5];
/* number of cpl delta bit segments */
		unsigned int cpldeltnseg;
/* coupling delta bit allocation offset */
			unsigned short cpldeltoffst[8];
/* coupling delta bit allocation length */
			unsigned short cpldeltlen[8];
/* coupling delta bit allocation length */
			unsigned short cpldeltba[8];
/* number of delta bit segments */
		unsigned short deltnseg[5];
/* fbw delta bit allocation offset */
			unsigned short deltoffst[5][8];
/* fbw delta bit allocation length */
			unsigned short deltlen[5][8];
/* fbw delta bit allocation length */
			unsigned short deltba[5][8];

/* skip length exists */
	unsigned int skiple;
/* skip length */
		unsigned int skipl;

/* channel mantissas */
	unsigned short chmant[5][256];

/* coupling mantissas */
	unsigned short cplmant[256];

/* coupling mantissas */
	unsigned short lfemant[7];

/*  -- Information not in the bitstream, but derived thereof  -- */

/* Number of coupling sub-bands */
	unsigned int ncplsubnd;

/* Number of combined coupling sub-bands
 * Derived from ncplsubnd and cplbndstrc */
	unsigned int ncplbnd;

/* Number of exponent groups by channel
 * Derived from strmant, endmant */
	unsigned short nchgrps[5];

/* Number of coupling exponent groups
 * Derived from cplbegf, cplendf, cplexpstr */
	unsigned int ncplgrps;
			
/* End mantissa numbers of fbw channels */
	unsigned short endmant[5];

/* Start and end mantissa numbers for the coupling channel */
	unsigned int cplstrtmant;
	unsigned int cplendmant;

/* Decoded exponent info */
	unsigned short fbw_exp[5][256];
	unsigned short cpl_exp[256];
	unsigned short lfe_exp[7];

/* Bit allocation pointer results */
	unsigned short fbw_bap[5][256];
	//FIXME figure out exactly how many entries there should be (253-37?) 
	unsigned short cpl_bap[256];
	unsigned short lfe_bap[7];
	
	unsigned long	magic3;
} mpeg2_ac3audblk_t;

// Bit allocation data
typedef struct
{
	int sdecay;
	int fdecay;
	int sgain;
	int dbknee;
	int floor;
	short psd[256];
	short bndpsd[256];
	short excite[256];
	short mask[256];
} mpeg2_ac3_bitallocation_t;

// Mantissa data
typedef struct
{
	unsigned short m_1[3];
	unsigned short m_2[3];
	unsigned short m_4[2];
	unsigned short m_1_pointer;
	unsigned short m_2_pointer;
	unsigned short m_4_pointer;
} mpeg2_ac3_mantissa_t;

#endif
