#include "../mpeg2private.h"
#include "mpeg2video.h"
#include "vlc.h"

#include <stdio.h>
#include <stdlib.h>

#ifdef HAVE_MMX
static unsigned long long MMX_128 = 0x80008000800080LL;
#endif

int mpeg2video_get_cbp(mpeg2_slice_t *slice)
{
  	int code;

  	if((code = mpeg2slice_showbits(slice->slice_buffer, 9)) >= 128)
	{
    	code >>= 4;
    	mpeg2slice_flushbits(slice->slice_buffer, mpeg2_CBPtab0[code].len);
    	return mpeg2_CBPtab0[code].val;
  	}

  	if(code >= 8)
	{
    	code >>= 1;
    	mpeg2slice_flushbits(slice->slice_buffer, mpeg2_CBPtab1[code].len);
    	return mpeg2_CBPtab1[code].val;
  	}

  	if(code < 1)
	{
    	fprintf(stderr,"mpeg2video_get_cbp: invalid coded_block_pattern code\n");
    	slice->fault = 1;
    	return 0;
  	}

  	mpeg2slice_flushbits(slice->slice_buffer, mpeg2_CBPtab2[code].len);
  	return mpeg2_CBPtab2[code].val;
}


/* set block to zero */
int mpeg2video_clearblock(mpeg2_slice_t *slice, int comp, int size)
{
	slice->sparse[comp] = 1;
	memset(slice->block[comp], 0, sizeof(short) * 64 * size);
	return 0;
}

int mpeg2video_getdclum(mpeg2_slice_buffer_t *slice_buffer)
{
	int code, size, val;

/* decode length */
	code = mpeg2slice_showbits(slice_buffer, 5);

	if(code < 31)
	{
    	size = mpeg2_DClumtab0[code].val;
    	mpeg2slice_flushbits(slice_buffer, mpeg2_DClumtab0[code].len);
	}
	else 
	{
    	code = mpeg2slice_showbits(slice_buffer, 9) - 0x1f0;
    	size = mpeg2_DClumtab1[code].val;
    	mpeg2slice_flushbits(slice_buffer, mpeg2_DClumtab1[code].len);
	}

	if(size == 0) val = 0;
	else 
	{
    	val = mpeg2slice_getbits(slice_buffer, size);
    	if((val & (1 << (size - 1))) == 0)  val -= (1 << size) - 1;
	}

	return val;
}


int mpeg2video_getdcchrom(mpeg2_slice_buffer_t *slice_buffer)
{
	int code, size, val;

/* decode length */
	code = mpeg2slice_showbits(slice_buffer, 5);

	if(code<31)
	{
    	size = mpeg2_DCchromtab0[code].val;
    	mpeg2slice_flushbits(slice_buffer, mpeg2_DCchromtab0[code].len);
	}
	else 
	{
    	code = mpeg2slice_showbits(slice_buffer, 10) - 0x3e0;
    	size = mpeg2_DCchromtab1[code].val;
    	mpeg2slice_flushbits(slice_buffer, mpeg2_DCchromtab1[code].len);
	}

	if(size == 0) val = 0;
	else 
	{
      val = mpeg2slice_getbits(slice_buffer, size);
      if((val & (1 << (size - 1))) == 0) val -= (1 << size) - 1;
	}

	return val;
}


// decode one intra coded MPEG-1 block

int mpeg2video_getintrablock(mpeg2_slice_t *slice, 
		mpeg2_layerdata_t *layerdata, 
		int comp, 
		int dc_dct_pred[])
{
	register int val, i, j, sign;
	unsigned int code;
	mpeg2_DCTtab_t *tab = 0;
	short *bp = slice->block[comp];

/* decode DC coefficients */
  	if(comp < 4)         
  		bp[0] = (dc_dct_pred[0] += mpeg2video_getdclum(slice->slice_buffer)) << 3;
  	else 
  	if(comp == 4)   
  		bp[0] = (dc_dct_pred[1] += mpeg2video_getdcchrom(slice->slice_buffer)) << 3;
	else                
  		bp[0] = (dc_dct_pred[2] += mpeg2video_getdcchrom(slice->slice_buffer)) << 3;

#ifdef HAVE_MMX
  	bp[0] <<= 4;
#endif

  	if(slice->fault) return 1;

/* decode AC coefficients */
  	for(i = 1; ; i++)
	{
    	code = mpeg2slice_showbits(slice->slice_buffer, 32);
    	code >>= 16;
    	if(code >= 16384)
			tab = &mpeg2_DCTtabnext[(code >> 12) - 4];
    	else 
		if(code >= 1024) tab = &mpeg2_DCTtab0[(code >> 8) - 4];
    	else 
		if(code >= 512) tab = &mpeg2_DCTtab1[(code >> 6) - 8];
    	else 
		if(code >= 256) tab = &mpeg2_DCTtab2[(code >> 4) - 16];
    	else 
		if(code >= 128) tab = &mpeg2_DCTtab3[(code >> 3) - 16];
    	else 
		if(code >= 64) tab = &mpeg2_DCTtab4[(code >> 2) - 16];
    	else 
		if(code >= 32) tab = &mpeg2_DCTtab5[(code >> 1) - 16];
    	else 
		if(code >= 16) tab = &mpeg2_DCTtab6[code - 16];
    	else 
		{
    	  	fprintf(stderr, "mpeg2video_getintrablock: invalid Huffman code\n");
    	  	slice->fault = 1;
    	  	return;
    	}

    	mpeg2slice_flushbits(slice->slice_buffer, tab->len);

    	if(tab->run == 64) break;  // end_of_block

    	if(tab->run == 65)
		{         
// escape
    		i += mpeg2slice_getbits(slice->slice_buffer, 6);

    		if((val = mpeg2slice_getbits(slice->slice_buffer, 8)) == 0) 
				val = mpeg2slice_getbits(slice->slice_buffer, 8);
    		else 
			if(val == 128)         
				val = mpeg2slice_getbits(slice->slice_buffer, 8) - 256;
    		else 
			if(val > 128)          
				val -= 256;

    		if((sign = (val < 0)) != 0) val= -val;
    	}
    	else 
		{
    		i += tab->run;
    		val = tab->level;
    		sign = mpeg2slice_getbits(slice->slice_buffer, 1);
    	}

    	j = mpeg2_zig_zag_scan[i];

#ifdef HAVE_MMX
    	val = (val * slice->quant_scale * layerdata->intra_quantizer_matrix[j]) << 1;
    	val = (val - 16) | 16;
#else
    	val = (val * slice->quant_scale * layerdata->intra_quantizer_matrix[j]) >> 3;
    	val = (val - 1) | 1;
#endif

    	bp[j] = sign ? -val : val;
	}

	if(j != 0) 
	{
/* not a layerdata->sparse matrix ! */
       slice->sparse[comp] = 0;
	}
	return 0;
}


// decode one non-intra coded MPEG-1 block

int mpeg2video_getinterblock(mpeg2_slice_t *slice, 
		mpeg2_layerdata_t *layerdata, 
		int comp)
{
	register int val, i, j, sign;
	unsigned int code;
	mpeg2_DCTtab_t *tab; 
	short *bp = slice->block[comp];

/* decode AC coefficients */
	for(i = 0; ; i++)
	{
    	code = mpeg2slice_showbits(slice->slice_buffer, 32);
    	code >>= 16;
    	if(code >= 16384)
		{
    	    if(i == 0) 
				tab = &mpeg2_DCTtabfirst[(code >> 12) - 4];
    	    else      
				tab = &mpeg2_DCTtabnext[(code >> 12) - 4];
    	}
    	else 
		if(code >= 1024) tab = &mpeg2_DCTtab0[(code >> 8) - 4];
    	else 
		if(code >= 512)  tab = &mpeg2_DCTtab1[(code >> 6) - 8];
    	else 
		if(code >= 256)  tab = &mpeg2_DCTtab2[(code >> 4) - 16];
    	else 
		if(code >= 128)  tab = &mpeg2_DCTtab3[(code >> 3) - 16];
    	else 
		if(code >= 64)   tab = &mpeg2_DCTtab4[(code >> 2) - 16];
    	else 
		if(code >= 32)   tab = &mpeg2_DCTtab5[(code >> 1) - 16];
    	else 
		if(code >= 16)   tab = &mpeg2_DCTtab6[code - 16];
    	else 
		{
    		fprintf(stderr,"mpeg2video_getinterblock: invalid Huffman code\n");
    		slice->fault = 1;
    		return 1;
    	}

    	mpeg2slice_flushbits(slice->slice_buffer, tab->len);

// end of block
    	if(tab->run == 64)
    	   break;   

    	if(tab->run == 65)
		{          
// escape 
    		i += mpeg2slice_getbits(slice->slice_buffer, 6);
    		if((val = mpeg2slice_getbits(slice->slice_buffer, 8)) == 0) 
				val = mpeg2slice_getbits(slice->slice_buffer, 8);
    		else 
			if(val == 128)  
				val = mpeg2slice_getbits(slice->slice_buffer, 8) - 256;
    		else 
			if(val > 128) 
				val -= 256;

    		if((sign = (val < 0)) != 0) val = -val;
    	}
    	else 
		{
    		i += tab->run;
    		val = tab->level;
    		sign = mpeg2slice_getbits(slice->slice_buffer, 1);
    	}

    	j = mpeg2_zig_zag_scan[i];
#ifdef HAVE_MMX
    	val = (((val << 1)+1) * slice->quant_scale * layerdata->non_intra_quantizer_matrix[j]);
    	val = (val - 16) | 16;
#else
    	val = (((val << 1)+1) * slice->quant_scale * layerdata->non_intra_quantizer_matrix[j]) >> 4;
    	val = (val - 1) | 1;
#endif
    	bp[j] = sign ? -val : val;
	}

	if(j != 0) 
	{
/* not a sparse matrix ! */
       slice->sparse[comp] = 0;
	}
	return 0;
}


// decode one intra coded MPEG-2 block
int mpeg2video_getmpg2intrablock(mpeg2_slice_t *slice, 
		mpeg2video_t *video, 
		mpeg2_layerdata_t *layerdata, 
		int comp, 
		int dc_dct_pred[])
{
	register int val, i, j, sign, nc;
	unsigned int code;
	mpeg2_DCTtab_t *tab;
	short *bp;
	int *qmat;

/* with data partitioning, data always goes to base layer */
  	bp = slice->block[comp];

  	qmat = (comp < 4 || video->chroma_format == CHROMA420)
         ? layerdata->intra_quantizer_matrix
         : layerdata->chroma_intra_quantizer_matrix;

/* decode DC coefficients */
	if(comp < 4)           
		val = (dc_dct_pred[0] += mpeg2video_getdclum(slice->slice_buffer));
	else 
	if((comp & 1) == 0) 
		val = (dc_dct_pred[1] += mpeg2video_getdcchrom(slice->slice_buffer));
	else                  
		val = (dc_dct_pred[2] += mpeg2video_getdcchrom(slice->slice_buffer));

  	if(slice->fault) return;
#ifdef HAVE_MMX
  	bp[0] = val << (7 - video->dc_prec);
#else
  	bp[0] = val << (3 - video->dc_prec);
#endif
  	nc = 0;

/* decode AC coefficients */
  	for(i = 1; ; i++)
	{
    	code = mpeg2slice_showbits(slice->slice_buffer, 16);
    	if(code >= 16384 && !video->intravlc)
			tab = &mpeg2_DCTtabnext[(code >> 12) - 4];
    	else 
		if(code >= 1024)
		{
    		if(video->intravlc) 
				tab = &mpeg2_DCTtab0a[(code >> 8) - 4];
    		else 
				tab = &mpeg2_DCTtab0[(code >> 8) - 4];
    	}
    	else 
		if(code >= 512)
		{
    		if(video->intravlc)     
		  	  	tab = &mpeg2_DCTtab1a[(code >> 6) - 8];
    		else              
				tab = &mpeg2_DCTtab1[(code >> 6) - 8];
    	}
    	else 
		if(code >= 256) tab = &mpeg2_DCTtab2[(code >> 4) - 16];
    	else 
		if(code >= 128) tab = &mpeg2_DCTtab3[(code >> 3) - 16];
    	else 
		if(code >= 64)  tab = &mpeg2_DCTtab4[(code >> 2) - 16];
    	else 
		if(code >= 32)  tab = &mpeg2_DCTtab5[(code >> 1) - 16];
    	else 
		if(code >= 16)  tab = &mpeg2_DCTtab6[code - 16];
    	else 
		{
    		fprintf(stderr,"mpeg2video_getmpg2intrablock: invalid Huffman code\n");
    		slice->fault = 1;
    		return 1;
    	}

    	mpeg2slice_flushbits(slice->slice_buffer, tab->len);

// end_of_block
    	if(tab->run == 64)
    	   	break; 

    	if(tab->run == 65)
		{        
// escape
    	  	i += mpeg2slice_getbits(slice->slice_buffer, 6);

    	  	val = mpeg2slice_getbits(slice->slice_buffer, 12);
    	  	if((val & 2047) == 0)
			{
        		fprintf(stderr,"mpeg2video_getmpg2intrablock: invalid signed_level (escape)\n");
        		slice->fault = 1;
        		return;
    	  	}
    	  	if((sign = (val >= 2048)) != 0) val = 4096 - val;
    	}
    	else 
		{
    		i += tab->run;
    		val = tab->level;
    		sign = mpeg2slice_getbits(slice->slice_buffer, 1);
    	}

    	j = (layerdata->altscan ? mpeg2_alternate_scan : mpeg2_zig_zag_scan)[i];

#ifdef HAVE_MMX
    	val = (val * slice->quant_scale * qmat[j]);
#else
    	val = (val * slice->quant_scale * qmat[j]) >> 4;
#endif

    	bp[j] = sign ? -val : val;
    	nc++;
	}

	if(j != 0) 
	{
/* not a sparse matrix ! */
    	 slice->sparse[comp] = 0;
	}
	return 1;
}


// decode one non-intra coded MPEG-2 block

int mpeg2video_getmpg2interblock(mpeg2_slice_t *slice, 
		mpeg2video_t *video, 
		mpeg2_layerdata_t *layerdata, 
		int comp)
{
	register int val, i, j, sign, nc;
	unsigned int code;
	mpeg2_DCTtab_t *tab;
	short *bp;
	int *qmat;

/* with data partitioning, data always goes to base layer */
  	bp = slice->block[comp];

  	qmat = (comp < 4 || video->chroma_format == CHROMA420)
         ? layerdata->non_intra_quantizer_matrix
         : layerdata->chroma_non_intra_quantizer_matrix;

  	nc = 0;

/* decode AC coefficients */
  	for(i = 0; ; i++)
	{
    	code = mpeg2slice_showbits(slice->slice_buffer, 16);
    	if(code >= 16384)
		{
    	  if(i == 0) tab = &mpeg2_DCTtabfirst[(code >> 12) - 4];
    	  else      tab = &mpeg2_DCTtabnext[(code >> 12) - 4];
    	}
    	else 
		if(code >= 1024) tab = &mpeg2_DCTtab0[(code >> 8) - 4];
    	else 
		if(code >= 512)  tab = &mpeg2_DCTtab1[(code >> 6) - 8];
    	else 
		if(code >= 256)  tab = &mpeg2_DCTtab2[(code >> 4) - 16];
    	else 
		if(code >= 128)  tab = &mpeg2_DCTtab3[(code >> 3) - 16];
    	else 
		if(code >= 64)   tab = &mpeg2_DCTtab4[(code >> 2) - 16];
    	else 
		if(code >= 32)   tab = &mpeg2_DCTtab5[(code >> 1) - 16];
    	else 
		if(code >= 16)   tab = &mpeg2_DCTtab6[code - 16];
    	else 
		{
    		fprintf(stderr, "mpeg2video_getmpg2interblock: invalid Huffman code\n");
    		slice->fault = 1;
    		return;
    	}

    	mpeg2slice_flushbits(slice->slice_buffer, tab->len);

// end_of_block
    	if(tab->run == 64)
       		break;          

    	if(tab->run == 65)
		{                 
// escape
    		i += mpeg2slice_getbits(slice->slice_buffer, 6);
    		val = mpeg2slice_getbits(slice->slice_buffer, 12);
    		if((val & 2047) == 0)
			{
        		fprintf(stderr, "mpeg2video_getmpg2interblock: invalid signed_level (escape)\n");
        		slice->fault = 1;
        		return 1;
    		}
    		if((sign = (val >= 2048)) != 0) val = 4096 - val;
    	}
    	else 
		{
    		i += tab->run;
    		val = tab->level;
    		sign = mpeg2slice_getbits(slice->slice_buffer, 1);
    	}

    	j = (layerdata->altscan ? mpeg2_alternate_scan : mpeg2_zig_zag_scan)[i];
#ifdef HAVE_MMX
     	val = (((val << 1)+1) * slice->quant_scale * qmat[j]) >> 1;
#else
     	val = (((val << 1)+1) * slice->quant_scale * qmat[j]) >> 5;
#endif

    	bp[j] = sign ? (-val) : val ;
    	nc++;
	}

	if(j != 0) 
	{
      	slice->sparse[comp] = 0;
	}
	return 0;
}


/* limit coefficients to -2048..2047 */

/* move/add 8x8-Block from block[comp] to refframe */

int mpeg2video_addblock(mpeg2_slice_t *slice, 
		mpeg2video_t *video, 
		mpeg2_layerdata_t *layerdata, 
		int comp, 
		int bx, 
		int by, 
		int dct_type, 
		int addflag)
{
	int cc, i, iincr;
	unsigned char *rfp;
	short *bp;
	int spar = slice->sparse[comp];
/* color component index */
  	cc = (comp < 4) ? 0 : (comp & 1) + 1; 

  	if(cc == 0)
	{   
/* luminance */
    	if(video->pict_struct == FRAME_PICTURE)
		{
      		if(dct_type)
			{
/* field DCT coding */
        		rfp = video->newframe[0] + 
              		video->coded_picture_width * (by + ((comp & 2) >> 1)) + bx + ((comp & 1) << 3);
        		iincr = (video->coded_picture_width << 1);
      		}
      		else
			{
/* frame DCT coding */
        		rfp = video->newframe[0] + 
             		video->coded_picture_width * (by + ((comp & 2) << 2)) + bx + ((comp & 1) << 3);
        		iincr = video->coded_picture_width;
      		}
		}
    	else 
		{
/* field picture */
      		rfp = video->newframe[0] + 
           		(video->coded_picture_width << 1) * (by + ((comp & 2) << 2)) + bx + ((comp & 1) << 3);
      		iincr = (video->coded_picture_width << 1);
    	}
 	}
  	else 
	{
/* chrominance */

/* scale coordinates */
    	if(video->chroma_format != CHROMA444) bx >>= 1;
    	if(video->chroma_format == CHROMA420) by >>= 1;
    	if(video->pict_struct == FRAME_PICTURE)
		{
    		if(dct_type && (video->chroma_format != CHROMA420))
			{
/* field DCT coding */
        		rfp = video->newframe[cc]
            		  + video->chrom_width * (by + ((comp & 2) >> 1)) + bx + (comp & 8);
        		iincr = (video->chrom_width << 1);
    		}
    		else 
			{
/* frame DCT coding */
        		rfp = video->newframe[cc]
            		  + video->chrom_width * (by + ((comp & 2) << 2)) + bx + (comp & 8);
        		iincr = video->chrom_width;
    		}
    	}
    	else 
		{
/* field picture */
    		rfp = video->newframe[cc]
            	  + (video->chrom_width << 1) * (by + ((comp & 2) << 2)) + bx + (comp & 8);
    		iincr = (video->chrom_width << 1);
    	}
  	}

  	bp = slice->block[comp];

	if(addflag)
	{
#ifdef HAVE_MMX
    	if(spar)
		{
       __asm__ __volatile__(
            "movq       (%2),  %%mm6\n"  /* 4 blockvals */
            "pxor       %%mm4, %%mm4\n"
            "punpcklwd  %%mm6, %%mm6\n"
            "punpcklwd  %%mm6, %%mm6\n"
            ".align 8\n"
            "1:"
                "movq       (%1),  %%mm0\n"     /* 8 rindex1 */
                "movq       %%mm0, %%mm2\n"
                "punpcklbw  %%mm4, %%mm0\n"
                "punpckhbw  %%mm4, %%mm2\n"
                "paddw      %%mm6, %%mm0\n"
                "paddw      %%mm6, %%mm2\n"

                "packuswb   %%mm2, %%mm0\n"
                "movq       %%mm0, (%1)\n"

                "leal       (%1, %3), %1\n"
              "loop       1b\n"
              :              /* scr   dest */
              : "c" (8),"r" (rfp), "r" (bp), "r" (iincr)
                );
    	  }
    	  else 
		  {
       __asm__ __volatile__(
         "pxor    %%mm4, %%mm4\n"

         ".align 8\n"
         "1:"
           "movq       (%2), %%mm0\n"   /* 8 rfp 0 1 2 3 4 5 6 7*/
           "movq       (%1), %%mm6\n"   /* 4 blockvals 0 1 2 3 */

           "movq       %%mm0, %%mm2\n"
           "movq       8(%1), %%mm5\n"  /* 4 blockvals 0 1 2 3 */
           "punpcklbw  %%mm4, %%mm0\n"  /* 0 2 4 6 */
           "punpckhbw  %%mm4, %%mm2\n"  /* 1 3 5 7 */

           "paddw      %%mm6, %%mm0\n"
           "paddw      %%mm5, %%mm2\n"
           "packuswb   %%mm2, %%mm0\n"

           "addl       $16, %1\n"
           "movq       %%mm0, (%2)\n"

           "leal       (%2,%3), %2\n"
         "loop       1b\n"
         :              /* scr   dest */
         : "c" (8),"r" (bp), "r" (rfp), "r" (iincr)
    	);
    	}
#else
		for(i = 0; i < 8; i++)
		{
    		rfp[0] = CLIP(bp[0] + rfp[0]);
    		rfp[1] = CLIP(bp[1] + rfp[1]);
    		rfp[2] = CLIP(bp[2] + rfp[2]);
    		rfp[3] = CLIP(bp[3] + rfp[3]);
    		rfp[4] = CLIP(bp[4] + rfp[4]);
    		rfp[5] = CLIP(bp[5] + rfp[5]);
    		rfp[6] = CLIP(bp[6] + rfp[6]);
    		rfp[7] = CLIP(bp[7] + rfp[7]);
    		rfp += iincr;
    		bp += 8;
		}
#endif
  	}
  	else 
  	{
#ifdef HAVE_MMX
    	if(spar) 
		{
            __asm__ __volatile__(
            "movd       (%2),           %%mm0\n"    // " 0 0 0  v1"
            "punpcklwd  %%mm0,          %%mm0\n"    // " 0 0 v1 v1"
            "punpcklwd  %%mm0,          %%mm0\n"
            "paddw      MMX_128,        %%mm0\n"
            "packuswb   %%mm0,          %%mm0\n"
            "leal       (%0,%1,2),      %%eax\n"

            "movq        %%mm0,         (%0, %1)\n"
            "movq        %%mm0,         (%%eax)\n"
            "leal        (%%eax,%1,2),  %0\n"
            "movq        %%mm0,         (%%eax, %1)\n"

            "movq        %%mm0,         (%0)\n"
            "leal        (%0,%1,2),     %%eax\n"
            "movq        %%mm0,         (%0, %1)\n"

            "movq        %%mm0,         (%%eax)\n"
            "movq        %%mm0,         (%%eax, %1)\n"
            :
            : "D" (rfp), "c" (iincr), "b" (bp)
            : "eax");
    	}
    	else 
		{

       __asm__ __volatile__(
            "movq        MMX_128,%%mm4\n"
            ".align 8\n"
            "1:"
              "movq      (%1),   %%mm0\n"
              "movq      8(%1),  %%mm1\n"
              "paddw     %%mm4,  %%mm0\n"

              "movq      16(%1), %%mm2\n"
              "paddw     %%mm4,  %%mm1\n"

              "movq      24(%1), %%mm3\n"
              "paddw     %%mm4,  %%mm2\n"

              "packuswb  %%mm1,  %%mm0\n"
              "paddw     %%mm4,  %%mm3\n"

              "addl $32, %1\n"
              "packuswb  %%mm3,  %%mm2\n"

              "movq   %%mm0, (%2)\n"

              "movq   %%mm2, (%2,%3)\n"

              "leal       (%2,%3,2), %2\n"
            "loop       1b\n"
            :
            : "c" (4), "r" (bp), "r" (rfp), "r" (iincr)
        );

//      set_block_mmx(rfp,bp,iincr);
    	}
#else
    	for(i = 0; i < 8; i++)
		{
    		rfp[0] = CLIP(bp[0] + 128);
    		rfp[1] = CLIP(bp[1] + 128);
    		rfp[2] = CLIP(bp[2] + 128);
    		rfp[3] = CLIP(bp[3] + 128);
    		rfp[4] = CLIP(bp[4] + 128);
    		rfp[5] = CLIP(bp[5] + 128);
    		rfp[6] = CLIP(bp[6] + 128);
    		rfp[7] = CLIP(bp[7] + 128);
    		rfp+= iincr;
    		bp += 8;
    	}
#endif
  	}
	return 0;
}

/* decode all macroblocks of the current picture */
int mpeg2video_get_macroblocks(mpeg2video_t *video, mpeg2_layerdata_t *layerdata, int framenum)
{
	unsigned int code;
	mpeg2_slice_buffer_t *slice_buffer; // Buffer being loaded
	int i;
	int current_buffer = 0;


// Load every slice into a buffer array
	video->total_slice_buffers = 0;
	while(mpeg2bits_showbits(video->vstream, 32) >= MPEG2_SLICE_MIN_START && 
		mpeg2bits_showbits(video->vstream, 32) <= MPEG2_SLICE_MAX_START && 
		!mpeg2bits_error(video->vstream))
	{
// Initialize the buffer
		if(current_buffer >= video->slice_buffers_initialized)
			mpeg2_new_slice_buffer(&(video->slice_buffers[video->slice_buffers_initialized++]));
		slice_buffer = &(video->slice_buffers[current_buffer]);
		slice_buffer->buffer_size = 0;
		slice_buffer->current_position = 0;
		slice_buffer->bits_size = 0;
		slice_buffer->done = 0;

// Read the slice into the buffer including the slice start code
		do
		{
// Expand buffer
			if(slice_buffer->buffer_allocation <= slice_buffer->buffer_size)
				mpeg2_expand_slice_buffer(slice_buffer);

// Load 1 char into buffer
			if(!mpeg2bits_error(video->vstream))
			{
				slice_buffer->data[slice_buffer->buffer_size++] = mpeg2bits_getbits(video->vstream, 8);
			}
		}while(!mpeg2bits_error(video->vstream) &&
			mpeg2bits_showbits(video->vstream, 24) != MPEG2_PACKET_START_CODE_PREFIX);

// Pad the buffer to get the last macroblock
		if(slice_buffer->buffer_allocation <= slice_buffer->buffer_size + 4)
			mpeg2_expand_slice_buffer(slice_buffer);

		slice_buffer->data[slice_buffer->buffer_size++] = 0;
		slice_buffer->data[slice_buffer->buffer_size++] = 0;
		slice_buffer->data[slice_buffer->buffer_size++] = 1;
		slice_buffer->data[slice_buffer->buffer_size++] = 0;

		if(!mpeg2bits_error(video->vstream))
			pthread_mutex_lock(&(slice_buffer->completion_lock));
		current_buffer++;
		video->total_slice_buffers++;
	}

// Run the slice decoders
	if(!mpeg2bits_error(video->vstream))
	{
		for(i = 0; i < video->total_slice_decoders; i++)
		{
			if(i == 0 && video->total_slice_decoders > 1)
			{
				video->slice_decoders[i].current_buffer = 0;
				video->slice_decoders[i].buffer_step = 1;
				video->slice_decoders[i].last_buffer = (video->total_slice_buffers - 1);
			}
			else
			if(i == 1)
			{
				video->slice_decoders[i].current_buffer = video->total_slice_buffers - 1;
				video->slice_decoders[i].buffer_step = -1;
				video->slice_decoders[i].last_buffer = 0;
			}
			else
			{
				video->slice_decoders[i].current_buffer = i;
				video->slice_decoders[i].buffer_step = 1;
				video->slice_decoders[i].last_buffer = video->total_slice_buffers - 1;
			}
			pthread_mutex_unlock(&(video->slice_decoders[i].input_lock));
		}
	}

// Wait for the slice decoders to finish
	if(!mpeg2bits_error(video->vstream))
	{
		for(i = 0; i < video->total_slice_buffers; i++)
		{
			pthread_mutex_lock(&(video->slice_buffers[i].completion_lock));
			pthread_mutex_unlock(&(video->slice_buffers[i].completion_lock));
		}
	}
	return 0;
}

// decode one frame or field picture

int mpeg2video_getpicture(mpeg2video_t *video, int framenum)
{
	int i, result = 0;
	mpeg2_t *file = video->file;
	if(video->pict_struct == FRAME_PICTURE && video->secondfield)
	{
/* recover from illegal number of field pictures */
    	video->secondfield = 0;
	}

// Get the slice decoders
	if(video->total_slice_decoders < file->cpus)
	{
		for(i = video->total_slice_decoders; i < file->cpus && i < MPEG2_MAX_CPUS; i++)
		{
			mpeg2_new_slice_decoder(video, &(video->slice_decoders[i]));
			video->slice_decoders[i].thread_number = i;
		}
		
		video->total_slice_decoders = file->cpus;
	}

  	for(i = 0; i < 3; i++)
	{
    	if(video->pict_type == B_TYPE)
		{
			video->newframe[i] = video->auxframe[i];
		}
    	else 
		{
    	  	if(!video->secondfield)
			{
// Swap refframes for interlaced
        		unsigned char* tmp = video->oldrefframe[i];
        		video->oldrefframe[i] = video->refframe[i];
        		video->refframe[i] = tmp;
    	  	}

    	 	video->newframe[i] = video->refframe[i];
    	}

    	if(video->pict_struct == BOTTOM_FIELD)
		{
// Only used if fields are in different pictures
    	    video->newframe[i] += (i == 0) ? video->coded_picture_width : video->chrom_width;
		}
	}

  	result |= mpeg2video_get_macroblocks(video, video->layerdata, framenum);

// Display the frame
	if(framenum > -1 && !result)
	{
    	if(video->pict_struct == FRAME_PICTURE || video->secondfield)
		{
     	  	if(video->pict_type == B_TYPE)
			{
 				mpeg2video_dither(video, video->auxframe);
			}
     	  	else
			{
 				mpeg2video_dither(video, video->oldrefframe);
			}
    	}
    	else 
		{
			mpeg2video_display_second_field(video);
		}
	}

  	if(video->pict_struct != FRAME_PICTURE) video->secondfield = !video->secondfield;
	return result;
}
