#ifndef MPEGVIDEO_H
#define MPEGVIDEO_H

#include "../bitstream.h"
#include "layerdata.h"
#include "../mpeg2private.inc"
#include "idct.h"
#include "slice.h"
#include "../timecode.h"

/* zig-zag scan */
extern unsigned char mpeg2_zig_zag_scan[64];

/* alternate scan */
extern unsigned char mpeg2_alternate_scan[64];

/* default intra quantization matrix */
extern unsigned char mpeg2_default_intra_quantizer_matrix[64];

/* Frame rate table must agree with the one in the encoder */
extern double mpeg2_frame_rate_Table[16];

/* non-linear quantization coefficient table */
extern unsigned char mpeg2_non_linear_mquant_table[32];

#define CHROMA420     1     // chroma_format
#define CHROMA422     2
#define CHROMA444     3

#define TOP_FIELD     1     // picture structure
#define BOTTOM_FIELD  2
#define FRAME_PICTURE 3

#define SEQ_ID        1     // extension start code IDs
#define DISP_ID       2
#define QUANT_ID      3
#define SEQSCAL_ID    5
#define PANSCAN_ID    7
#define CODING_ID     8
#define SPATSCAL_ID   9
#define TEMPSCAL_ID  10

#define ERROR (-1)

#define SC_NONE       0   // scalable_mode
#define SC_DP         1
#define SC_SPAT       2
#define SC_SNR        3
#define SC_TEMP       4

#define I_TYPE        1     // picture coding type
#define P_TYPE        2
#define B_TYPE        3
#define D_TYPE        4

#define MB_INTRA      1 	// macroblock type
#define MB_PATTERN    2
#define MB_BACKWARD   4
#define MB_FORWARD    8
#define MB_QUANT      16
#define MB_WEIGHT     32
#define MB_CLASS4     64

#define MC_FIELD      1     // motion_type
#define MC_FRAME      2
#define MC_16X8       2
#define MC_DMV        3

#define MV_FIELD      0     // mv_format
#define MV_FRAME      1

#define CLIP(x)  ((x) >= 0 ? ((x) < 255 ? (x) : 255) : 0)

// Statically allocate as little as possible so a fake video struct
// can be used for reading the GOP headers.

typedef struct
{
	void* file;
	void* track;

// ================================= new compression variables
	mpeg2_layerdata_t *layerdata;
	mpeg2_bits_t *vstream;
	int decoder_initted;
	unsigned char **output_rows;     // Output frame buffer supplied by user
	int output_w, output_h;          // Output dimensions
	int *x_table, *y_table;          // Location of every output pixel in the input
	int color_model;

	mpeg2_slice_t slice_decoders[MPEG2_MAX_CPUS];  // One slice decoder for every CPU
	int total_slice_decoders;                       // Total slice decoders in use
	mpeg2_slice_buffer_t slice_buffers[MPEG2_MAX_CPUS];   // Buffers for holding the slice data
	int total_slice_buffers;         // Total buffers in the array to be decompressed
	int slice_buffers_initialized;     // Total buffers initialized in the array
	pthread_mutex_t slice_lock;      // Lock slice array while getting the next buffer
	pthread_mutex_t test_lock;

	int blockreadsize;
	int framerate;
	long maxframe;         // Max value of frame num to read
	long framenum;         // Number of the frame used only by the encoder.
	unsigned char **last_src;    // Last frame dithered
	long last_number;       // Last framenum rendered
	int found_seqhdr;
	long bitrate;
	mpeg2_timecode_t gop_timecode;     // Timecode for the last GOP header read.
	long frames_per_gop;       // Frames per GOP after the first GOP.
	long first_gop_frames;     // Frames in the first GOP.
	long first_frame;     // Number of first frame stored in timecode
	long last_frame;      // Last frame in file

// ================================= Original compression variables
	int use_output;     // Transfer the frame to the output
// Malloced frame buffers.  2 refframes are swapped in and out for interlaced
// video while only 1 refframe is used for non-interlaced.
	unsigned char *oldrefframe[3], *refframe[3], *auxframe[3];
// Pointers to frame buffers.
	unsigned char *newframe[3];
	int horizontal_size, vertical_size, mb_width, mb_height;
	int coded_picture_width,  coded_picture_height;
	int chroma_format, chrom_width, chrom_height, blk_cnt;
	int pict_type;
	int forw_r_size, back_r_size;
	int full_forw, full_back;
	int verbose;
	int prog_seq;
	int h_forw_r_size, v_forw_r_size, h_back_r_size, v_back_r_size;
	int dc_prec, pict_struct, topfirst, frame_pred_dct, conceal_mv;
	int intravlc, repeatfirst, prog_frame;
	int secondfield;
	int stwc_table_index, llw, llh, hm, hn, vm, vn;
	int lltempref, llx0, lly0, llprog_frame, llfieldsel;
	unsigned char *llframe0[3], *llframe1[3];
	short *lltmp;
	char *llinputname;
	int sflag;
	int matrix_coefficients;
	int framerate_code;
	float frame_rate;
	int playedlastframe;
	mpeg2_idct_t *idct;
	long *cr_to_r, *cr_to_g, *cb_to_g, *cb_to_b;
	long *cr_to_r_ptr, *cr_to_g_ptr, *cb_to_g_ptr, *cb_to_b_ptr;
} mpeg2video_t;

#endif
