#include "asset.h"
#include "arender.h"
#include "audiodevice.h"
#include "file.h"
#include "mainwindow.h"

ARender::ARender(MainWindow *mwindow)
{
	this->mwindow = mwindow;
	reset_parameters();
}

ARender::~ARender()
{
}

int ARender::reset_parameters()
{
	audio = 0;
	file = 0;
	buffer = 0;
}

int ARender::arm_playback()
{
	file = new File;
	file->set_prebuffer(mwindow->prebuffer_size);
	file->set_processors(mwindow->smp ? 2: 1);
	audio = new AudioDevice(mwindow->asset);
	audio->set_software_sync(mwindow->software_sync);
	int result = file->try_to_open_file(mwindow->asset);
	synchronous = 1;

	if(!result)
	{
		startup_lock.lock();
		current_sample = mwindow->current_sample;
		file->set_audio_position(mwindow->current_sample);
		BUFFER_SIZE = audio->start_playback() * 8;
		bytes = mwindow->asset->bits / 8 * mwindow->asset->channels * BUFFER_SIZE;
		buffer = new char[bytes];

		return 0;
	}
	else
	{
		delete audio;
		delete file;
		reset_parameters();
		return 1;
	}
}

int ARender::stop_playback()
{
	if(audio) audio->interrupt_playback();
	end();    // end this thread
}

int ARender::close_playback()
{
	file->close_file();
	audio->stop_playback();
	delete file;
	delete buffer;
	delete audio;
	reset_parameters();
}

int ARender::wait_for_startup()
{
	startup_lock.lock();
	startup_lock.unlock();
}

int ARender::wait_for_completion()
{
	join();
	close_playback();
}

void ARender::run()
{
	startup_lock.unlock();
	while(current_sample < mwindow->total_samples)
	{
// Really need a threaded audio device.
		buffer_length = BUFFER_SIZE;
		if(current_sample + buffer_length > mwindow->total_samples)
			buffer_length = mwindow->total_samples - current_sample;

		if(buffer_length)
		{
			file->read_raw(buffer, buffer_length);
			audio->write_audio(buffer, buffer_length);
			current_sample += buffer_length;
		}
	}
// Flush the audio device
	audio->write_audio(0, 0);
}

