#ifndef FILE_H
#define FILE_H

#include <stdio.h>


#include "asset.inc"
#include "file.inc"
#include "filebase.inc"
#include "sema.h"


// ======================================= include file types here

// generic file opened by user
class File
{
public:
	File();
	~File();


// ===================================== start here
	int set_processors(int cpus);
	int set_prebuffer(long size);

// return 0 if success/found format
// return 1 if failure/file not found
// return 2 if need format
// return 3 if project
// format may be preset by set_format
	int try_to_open_file(Asset *asset);    

// close file
	int close_file();

// get length of file
	long get_audio_length();
	long get_video_length();

// get current position
	long get_audio_position();
	long get_video_position();

// set position in samples or frames
	int set_audio_position(long x);
	int set_video_position(long x);

// read the number of samples of raw data for all audio channels
// advances file pointer
// return 1 if failed
	int read_raw(char *buffer, long len);

// Can this file read directly into the frame buffer?
	int frame_buffer_copy_possible(int color_model);
	int read_frame_buffer(unsigned char **frame_buffer, int w, int h, int bitmap_color_model);
// read the frame at the current video position into a raw buffer of rows and columns
	int read_frame(unsigned char *frame);

// load video into ram
	int load_into_ram();

	int lock_read();
	int unlock_read();

	Asset *asset;
	FileBase *file; // virtual class for file type
	Sema read_lock, write_lock;
	int cpus;
	long prebuffer_size;
	long total_frames;
	long total_samples;
};

#endif
