#include "arender.h"
#include "asset.h"
#include "audiodevice.h"
#include "mainwindow.h"
#include "playbackengine.h"
#include "playbackscroll.h"
#include "vrender.h"

PlaybackEngine::PlaybackEngine(MainWindow *mwindow)
 : Thread()
{
	this->mwindow = mwindow;
	reset_parameters();
}

PlaybackEngine::~PlaybackEngine()
{
}

int PlaybackEngine::reset_parameters()
{
	arender = 0;
	vrender = 0;
	playing_back = 0;
	scroll = 0;
}

int PlaybackEngine::start_playback()
{
	if(!playing_back)
	{
		playing_back = 1;
		complete.lock();
		startup_lock.lock();
		int result = 0;
		start_position = mwindow->current_sample;

		if(mwindow->asset->video_data)
		{
			vrender = new VRender(mwindow);
			result = vrender->arm_playback();
		}

		if(mwindow->asset->audio_data && !result)
		{
			arender = new ARender(mwindow);
			result = arender->arm_playback();
		}
		
		scroll = new PlaybackScroll(mwindow);
		synchronous = 0;

		start();
		timer.update();
		wait_for_startup();
	}
}

int PlaybackEngine::stop_playback()
{
	if(playing_back)
	{
		if(arender) arender->stop_playback();
		if(vrender) vrender->stop_playback();
		scroll->stop_playback();
		complete.lock();
		complete.unlock();
	}
}

int PlaybackEngine::close_playback()
{
	if(playing_back)
	{
		if(arender) delete arender;
		if(vrender) delete vrender;
		delete scroll;
		reset_parameters();
		mwindow->gui->playbutton->update(1);
		playing_back = 0;
	}
}

int PlaybackEngine::wait_for_startup()
{
	startup_lock.lock();
	startup_lock.unlock();
	if(arender) arender->wait_for_startup();
	if(vrender) vrender->wait_for_startup();
	scroll->wait_for_startup();
}

long PlaybackEngine::current_position()
{
	if(playing_back)
	{
		long result;

		if(arender && arender->audio) 
			result = arender->audio->samples_rendered() + start_position;
		else
// try getting from a video device next only if every frame
		if(vrender && mwindow->every_frame) 
			result = vrender->current_position();
		else
// use elapsed time
		{
			result = timer.get_scaled_difference(mwindow->samplerate) + start_position;
		}

		return result;
	}
	else
		return -1;
}

void PlaybackEngine::run()
{
	startup_lock.unlock();
	scroll->start_playback();
	if(arender) arender->start();
	if(vrender) vrender->start();

	if(arender) arender->wait_for_completion();
	if(vrender) vrender->wait_for_completion();
	scroll->stop_playback();
	scroll->wait_for_completion();

	close_playback();
	complete.unlock();
}
