#include "settings.h"
#include "mainwindow.h"
#include "playbackengine.h"
#include "vrender.inc"
#include "vrender.h"


VRender::VRender(MainWindow *mwindow)
{
	this->mwindow = mwindow;
}

VRender::~VRender()
{
}

int VRender::arm_playback()
{
	startup_lock.lock();
	synchronous = 1;
	current_frame = (long)toframes(mwindow->current_sample, mwindow->samplerate, mwindow->frame_rate);
	interrupt = 0;
	mwindow->gui->canvas->start_video();
	return 0;
}

int VRender::stop_playback()
{
	interrupt = 1;
}

int VRender::close_playback()
{
	mwindow->gui->canvas->stop_video();
//	mwindow->flash_frame(current_frame);
	mwindow->flash_armed_frame();
}

int VRender::wait_for_startup()
{
	startup_lock.lock();
	startup_lock.unlock();
}

int VRender::wait_for_completion()
{
	join();
}

long VRender::current_position()
{
	return (long)((float)current_frame / mwindow->frame_rate * mwindow->samplerate);
}

void VRender::run()
{
	long current_sample;     // Time the frame was finished being decompressed.
	long start_sample, end_sample; // Absolute counts.
	int done = 0, skip;
	long delay;
	long total_frames = (long)toframes(mwindow->total_samples, mwindow->samplerate, mwindow->frame_rate);
	long skip_countdown = VRENDER_THRESHOLD;
	long delay_countdown = VRENDER_THRESHOLD;
	
	framerate_counter = 0;
	startup_lock.unlock();
	while(!interrupt && !done)
	{
// Update the framerate counter
		if(!framerate_counter)
		{
			mwindow->actual_framerate = (float)((int)mwindow->frame_rate)  / ((float)framerate_timer.get_difference() / 1000);
			if(mwindow->settingsmenu->thread->thread_running)
			{
				mwindow->settingsmenu->thread->update_framerate();
			}
			framerate_counter = (int)mwindow->frame_rate;
			framerate_timer.update();
		}
		framerate_counter--;
// Perform the most time consuming part of frame decompression now.
		mwindow->arm_frame(current_frame);

// Determine the delay until the frame needs to be shown.
		current_sample = (long)(mwindow->engine->current_position());
// latest sample at which the frame can be shown.
		end_sample = tosamples(current_frame, mwindow->samplerate, mwindow->frame_rate);
// earliest sample by which the frame needs to be shown.
		start_sample = tosamples(current_frame - 1, mwindow->samplerate, mwindow->frame_rate);

		if(end_sample < current_sample)
		{
// Frame rendered late.  Flash it now.
			mwindow->gui->lock_window();
			mwindow->flash_armed_frame();
			mwindow->gui->unlock_window();

			if(mwindow->every_frame)
			{
// User wants every frame.
				current_frame++;
			}
			else
			if(skip_countdown > 0)
			{
// Maybe just a freak.
				current_frame++;
				skip_countdown--;
			}
			else
			{
// Get the frames to skip.
				delay_countdown = VRENDER_THRESHOLD;
				skip = 1 + (long)toframes(current_sample, mwindow->samplerate, mwindow->frame_rate) - 
					(long)toframes(end_sample, mwindow->samplerate, mwindow->frame_rate);
				current_frame += skip;
			}
		}
		else
		{
// Frame rendered early or just in time.
			current_frame++;

			if(delay_countdown > 0)
			{
// Maybe just a freak
				delay_countdown--;
			}
			else
			{
				skip_countdown = VRENDER_THRESHOLD;
				if(start_sample > current_sample)
				{
// Came before the earliest sample so delay
					delay = (long)((float)(start_sample - current_sample) * 1000 / mwindow->samplerate);
					timer.delay(delay);
				}
				else
				{
// Came after the earliest sample so keep going
				}
			}

// Flash frame now.
			mwindow->gui->lock_window();
			mwindow->flash_armed_frame();
			mwindow->gui->unlock_window();
		}

		if(current_frame > total_frames) done = 1;
	}

	close_playback();
}
